# Copyright 2018 Huawei Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The share accesses api."""

import ast

import webob

from manila.api import common
from manila.api.openstack import wsgi
from manila.api.views import share_accesses as share_access_views
from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.lock import api as resource_locks
from manila import share


class ShareAccessesController(wsgi.Controller, wsgi.AdminActionsMixin):
    """The Share accesses API V2 controller for the OpenStack API."""

    resource_name = 'share_access_rule'
    _view_builder_class = share_access_views.ViewBuilder

    def __init__(self):
        super(ShareAccessesController, self).__init__()
        self.share_api = share.API()
        self.resource_locks_api = resource_locks.API()

    @wsgi.Controller.api_version('2.45')
    @wsgi.Controller.authorize('get')
    def show(self, req, id):
        """Return data about the given share access rule."""
        context = req.environ['manila.context']
        share_access = self._get_share_access(context, id)
        restricted = self._is_rule_restricted(context, id)
        if restricted:
            share_access['restricted'] = True
        return self._view_builder.view(req, share_access)

    def _is_rule_restricted(self, context, id):
        search_opts = {
            'resource_id': id,
            'resource_action': constants.RESOURCE_ACTION_SHOW,
            'resource_type': 'access_rule'
        }
        locks, count = self.resource_locks_api.get_all(
            context, search_opts, show_count=True)

        if count:
            return self.resource_locks_api.access_is_restricted(context,
                                                                locks[0])
        return False

    def _get_share_access(self, context, share_access_id):
        try:
            return self.share_api.access_get(context, share_access_id)
        except exception.NotFound:
            msg = _("Share access rule %s not found.") % share_access_id
            raise webob.exc.HTTPNotFound(explanation=msg)

    def _validate_search_opts(self, req, search_opts):
        """Check if search opts parameters are valid."""
        access_type = search_opts.get('access_type', None)
        access_to = search_opts.get('access_to', None)

        if access_type and access_type not in ['ip', 'user', 'cert', 'cephx']:
            raise exception.InvalidShareAccessType(type=access_type)

        # If access_to is present but access type is not, it gets tricky to
        # validate its content
        if access_to and not access_type:
            msg = _("'access_type' parameter must be provided when specifying "
                    "'access_to'.")
            raise exception.InvalidInput(reason=msg)

        if access_type and access_to:
            common.validate_access(access_type=access_type,
                                   access_to=access_to,
                                   enable_ceph=True,
                                   enable_ipv6=True)

        access_level = search_opts.get('access_level', None)
        if access_level and access_level not in constants.ACCESS_LEVELS:
            raise exception.InvalidShareAccessLevel(level=access_level)

    @wsgi.Controller.authorize('index')
    def _index(self, req, support_for_access_filters=False):
        """Returns the list of access rules for a given share."""
        context = req.environ['manila.context']
        search_opts = {}
        search_opts.update(req.GET)
        if 'share_id' not in search_opts:
            msg = _("The field 'share_id' has to be specified.")
            raise webob.exc.HTTPBadRequest(explanation=msg)
        share_id = search_opts.pop('share_id', None)

        if 'metadata' in search_opts:
            search_opts['metadata'] = ast.literal_eval(
                search_opts['metadata'])
        if support_for_access_filters:
            try:
                self._validate_search_opts(req, search_opts)
            except (exception.InvalidShareAccessLevel,
                    exception.InvalidShareAccessType) as e:
                raise webob.exc.HTTPBadRequest(explanation=e.msg)
        try:
            share = self.share_api.get(context, share_id)
        except exception.NotFound:
            msg = _("Share %s not found.") % share_id
            raise webob.exc.HTTPBadRequest(explanation=msg)
        access_rules = self.share_api.access_get_all(
            context, share, search_opts)
        rule_list = []
        for rule in access_rules:
            restricted = self._is_rule_restricted(context, rule['id'])
            rule['restricted'] = restricted
            if (('access_to' in search_opts or 'access_key' in search_opts)
                    and restricted):
                continue
            rule_list.append(rule)

        return self._view_builder.list_view(req, rule_list)

    @wsgi.Controller.api_version('2.45', '2.81')
    def index(self, req):
        return self._index(req)

    @wsgi.Controller.api_version('2.82')
    def index(self, req): # pylint: disable=function-redefined  # noqa F811
        return self._index(req, support_for_access_filters=True)


def create_resource():
    return wsgi.Resource(ShareAccessesController())



# Copyright (c) 2023 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
PowerFlex specific NAS backend plugin.
"""
from oslo_config import cfg
from oslo_log import log
from oslo_utils import units

from manila.common import constants as const
from manila import exception
from manila.i18n import _
from manila.share.drivers.dell_emc.plugins import base as driver
from manila.share.drivers.dell_emc.plugins.powerflex import (
    object_manager as manager)

"""Version history:
    1.0 - Initial version
"""

VERSION = "1.0"

CONF = cfg.CONF

LOG = log.getLogger(__name__)

POWERFLEX_OPTS = [
    cfg.StrOpt('powerflex_storage_pool',
               help='Storage pool used to provision NAS.'),
    cfg.StrOpt('powerflex_protection_domain',
               help='Protection domain to use.'),
    cfg.StrOpt('dell_nas_backend_host',
               help='Dell NAS backend hostname or IP address.'),
    cfg.IntOpt('dell_nas_backend_port',
               default=443,
               help='Port number to use with the Dell NAS backend.'),
    cfg.StrOpt('dell_nas_server',
               help='Root directory or NAS server which owns the shares.'),
    cfg.StrOpt('dell_nas_login',
               help='User name for the Dell NAS backend.'),
    cfg.StrOpt('dell_nas_password',
               secret=True,
               help='Password for the Dell NAS backend.')

]


class PowerFlexStorageConnection(driver.StorageConnection):
    """Implements PowerFlex specific functionality for Dell Manila driver."""

    def __init__(self, *args, **kwargs):
        """Do initialization"""

        LOG.debug('Invoking base constructor for Manila \
                  Dell PowerFlex SDNAS Driver.')
        super(PowerFlexStorageConnection,
              self).__init__(*args, **kwargs)

        LOG.debug('Setting up attributes for Manila \
                  Dell PowerFlex SDNAS Driver.')
        if 'configuration' in kwargs:
            kwargs['configuration'].append_config_values(POWERFLEX_OPTS)

        self.manager = None
        self.server = None
        self._username = None
        self._password = None
        self._server_url = None
        self._root_dir = None
        self._verify_ssl_cert = None
        self._shares = {}
        self.verify_certificate = None
        self.certificate_path = None
        self.export_path = None

        self.driver_handles_share_servers = False

        self.reserved_percentage = None
        self.reserved_snapshot_percentage = None
        self.reserved_share_extend_percentage = None
        self.max_over_subscription_ratio = None

    def connect(self, dell_share_driver, context):
        """Connects to Dell PowerFlex SDNAS server."""
        LOG.debug('Reading configuration parameters for Manila \
                  Dell PowerFlex SDNAS Driver.')
        config = dell_share_driver.configuration
        get_config_value = config.safe_get
        self.verify_certificate = get_config_value("dell_ssl_cert_verify")
        self.rest_ip = get_config_value("dell_nas_backend_host")
        self.rest_port = get_config_value("dell_nas_backend_port")
        self.nas_server = get_config_value("dell_nas_server")
        self.storage_pool = get_config_value("powerflex_storage_pool")
        self.protection_domain = get_config_value(
            "powerflex_protection_domain")
        self.rest_username = get_config_value("dell_nas_login")
        self.rest_password = get_config_value("dell_nas_password")
        if self.verify_certificate:
            self.certificate_path = get_config_value(
                "dell_ssl_certificate_path")
        if not all([self.rest_ip,
                    self.rest_username,
                    self.rest_password]):
            message = _("REST server IP, username and password"
                        " must be specified.")
            raise exception.BadConfigurationException(reason=message)
        # validate certificate settings
        if self.verify_certificate and not self.certificate_path:
            message = _("Path to REST server's certificate must be specified.")
            raise exception.BadConfigurationException(reason=message)

        LOG.debug('Initializing Dell PowerFlex SDNAS Layer.')
        self.host_url = ("https://%(server_ip)s:%(server_port)s" %
                         {
                             "server_ip": self.rest_ip,
                             "server_port": self.rest_port})
        LOG.info("REST server IP: %(ip)s, port: %(port)s, "
                 "username: %(user)s. Verify server's certificate: "
                 "%(verify_cert)s.",
                 {
                     "ip": self.rest_ip,
                     "port": self.rest_port,
                     "user": self.rest_username,
                     "verify_cert": self.verify_certificate,
                 })

        self.manager = manager.StorageObjectManager(self.host_url,
                                                    self.rest_username,
                                                    self.rest_password,
                                                    self.export_path,
                                                    self.certificate_path,
                                                    self.verify_certificate)

        # configuration for share status update
        self.reserved_percentage = config.safe_get(
            'reserved_share_percentage')
        if self.reserved_percentage is None:
            self.reserved_percentage = 0

        self.reserved_snapshot_percentage = config.safe_get(
            'reserved_share_from_snapshot_percentage')
        if self.reserved_snapshot_percentage is None:
            self.reserved_snapshot_percentage = self.reserved_percentage

        self.reserved_share_extend_percentage = config.safe_get(
            'reserved_share_extend_percentage')
        if self.reserved_share_extend_percentage is None:
            self.reserved_share_extend_percentage = self.reserved_percentage

        self.max_over_subscription_ratio = config.safe_get(
            'max_over_subscription_ratio')

    def create_share(self, context, share, share_server):
        """Is called to create a share."""
        LOG.debug(f'Creating {share["share_proto"]} share.')
        location = self._create_nfs_share(share)

        return location

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        """Is called to create a share from an existing snapshot."""
        raise NotImplementedError()

    def allow_access(self, context, share, access, share_server):
        """Is called to allow access to a share."""
        raise NotImplementedError()

    def check_for_setup_error(self):
        """Is called to check for setup error."""

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Is called to update share access."""
        LOG.debug(f'Updating access to {share["share_proto"]} share.')
        return self._update_nfs_access(share, access_rules)

    def create_snapshot(self, context, snapshot, share_server):
        """Is called to create snapshot."""
        export_name = snapshot['share_name']
        LOG.debug(f'Retrieving filesystem ID for share {export_name}')
        filesystem_id = self.manager.get_fsid_from_export_name(export_name)
        LOG.debug(f'Retrieving snapshot ID for filesystem {filesystem_id}')
        snapshot_id = self.manager.create_snapshot(snapshot['name'],
                                                   filesystem_id)
        if snapshot_id:
            LOG.info("Snapshot %(id)s successfully created.",
                     {'id': snapshot['id']})

    def delete_snapshot(self, context, snapshot, share_server):
        """Is called to delete snapshot."""
        snapshot_name = snapshot['name']
        filesystem_id = self.manager.get_fsid_from_snapshot_name(snapshot_name)
        LOG.debug(f'Retrieving filesystem ID for snapshot {snapshot_name}')
        snapshot_deleted = self.manager.delete_filesystem(filesystem_id)
        if not snapshot_deleted:
            message = (
                _('Failed to delete snapshot "%(snapshot)s".') %
                {'snapshot': snapshot['name']})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        else:
            LOG.info("Snapshot %(id)s successfully deleted.",
                     {'id': snapshot['id']})

    def delete_share(self, context, share, share_server):
        """Is called to delete a share."""
        LOG.debug(f'Deleting {share["share_proto"]} share.')
        self._delete_nfs_share(share)

    def deny_access(self, context, share, access, share_server):
        """Is called to deny access to a share."""
        raise NotImplementedError()

    def ensure_share(self, context, share, share_server):
        """Is called to ensure a share is exported."""

    def extend_share(self, share, new_size, share_server=None):
        """Is called to extend a share."""
        # Converts the size from GiB to Bytes
        new_size_in_bytes = new_size * units.Gi
        LOG.debug(f"Extending {share['name']} to {new_size}GiB")
        filesystem_id = self.manager.get_filesystem_id(share['name'])
        self.manager.extend_export(filesystem_id,
                                   new_size_in_bytes)

    def setup_server(self, network_info, metadata=None):
        """Is called to set up a share server.

        Requires driver_handles_share_servers to be True.
        """
        raise NotImplementedError()

    def teardown_server(self, server_details, security_services=None):
        """Is called to teardown a share server.

        Requires driver_handles_share_servers to be True.
        """
        raise NotImplementedError()

    def _create_nfs_share(self, share):
        """Creates an NFS share.

        In PowerFlex, an export (share) belongs to a filesystem.
        This function creates a filesystem and an export.
        """
        LOG.debug(f'Retrieving Storage Pool ID for {self.storage_pool}')
        storage_pool_id = self.manager.get_storage_pool_id(
            self.protection_domain,
            self.storage_pool)
        nas_server_id = self.manager.get_nas_server_id(self.nas_server)
        LOG.debug(f"Creating filesystem {share['name']}")
        size_in_bytes = share['size'] * units.Gi
        filesystem_id = self.manager.create_filesystem(storage_pool_id,
                                                       self.nas_server,
                                                       share['name'],
                                                       size_in_bytes)
        if not filesystem_id:
            message = {
                _('The requested NFS export "%(export)s"'
                    ' was not created.') %
                {'export': share['name']}}
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)

        LOG.debug(f"Creating export {share['name']}")
        export_id = self.manager.create_nfs_export(filesystem_id,
                                                   share['name'])
        if not export_id:
            message = (
                _('The requested NFS export "%(export)s"'
                    ' was not created.') %
                {'export': share['name']})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        file_interfaces = self.manager.get_nas_server_interfaces(
            nas_server_id)
        export_path = self.manager.get_nfs_export_name(export_id)
        locations = self._get_nfs_location(file_interfaces,
                                           export_path)
        return locations

    def _delete_nfs_share(self, share):
        """Deletes a filesystem and its associated export."""
        filesystem_id = self.manager.get_filesystem_id(share['name'])
        LOG.debug(f"Retrieving filesystem ID for filesystem {share['name']}")
        if filesystem_id is None:
            message = ('Attempted to delete NFS export "%s",'
                       ' but the export does not appear to exist.')
            LOG.warning(message, share['name'])
        else:
            LOG.debug(f"Deleting filesystem ID {filesystem_id}")
            share_deleted = self.manager.delete_filesystem(filesystem_id)
            if not share_deleted:
                message = (
                    _('Failed to delete NFS export "%(export)s".') %
                    {'export': share['name']})
                LOG.error(message)
                raise exception.ShareBackendException(msg=message)

    def _update_nfs_access(self, share, access_rules):
        """Updates access rules for NFS share type."""
        nfs_rw_ips = set()
        nfs_ro_ips = set()
        access_updates = {}

        for rule in access_rules:
            if rule['access_type'].lower() != 'ip':
                message = (_("Only IP access type currently supported for "
                             "NFS. Share provided %(share)s with rule type "
                             "%(type)s") % {'share': share['display_name'],
                                            'type': rule['access_type']})
                LOG.error(message)
                access_updates.update({rule['access_id']: {'state': 'error'}})

            else:
                if rule['access_level'] == const.ACCESS_LEVEL_RW:
                    nfs_rw_ips.add(rule['access_to'])
                elif rule['access_level'] == const.ACCESS_LEVEL_RO:
                    nfs_ro_ips.add(rule['access_to'])
                access_updates.update({rule['access_id']: {'state': 'active'}})

        share_id = self.manager.get_nfs_export_id(share['name'])
        share_updated = self.manager.set_export_access(share_id,
                                                       nfs_rw_ips,
                                                       nfs_ro_ips)
        if not share_updated:
            message = (
                _('Failed to update NFS access rules for "%(export)s".') %
                {'export': share['display_name']})
            LOG.error(message)
            raise exception.ShareBackendException(msg=message)
        return access_updates

    def update_share_stats(self, stats_dict):
        """Retrieve stats info from share."""
        stats_dict['driver_version'] = VERSION
        stats_dict['storage_protocol'] = 'NFS'
        stats_dict['create_share_from_snapshot_support'] = False
        stats_dict['pools'] = []
        storage_pool_id = self.manager.get_storage_pool_id(
            self.protection_domain,
            self.storage_pool
            )
        total = free = used = provisioned = 0
        statistic = self.manager.get_storage_pool_statistic(storage_pool_id)
        if statistic:
            total = statistic.get('maxCapacityInKb') // units.Mi
            free = statistic.get('netUnusedCapacityInKb') // units.Mi
            used = statistic.get('capacityInUseInKb') // units.Mi
            provisioned = statistic.get('primaryVacInKb') // units.Mi
        pool_stat = {
            'pool_name': self.storage_pool,
            'thin_provisioning': True,
            'total_capacity_gb': total,
            'free_capacity_gb': free,
            'allocated_capacity_gb': used,
            'provisioned_capacity_gb': provisioned,
            'qos': False,
            'reserved_percentage': self.reserved_percentage,
            'reserved_snapshot_percentage':
                self.reserved_snapshot_percentage,
            'reserved_share_extend_percentage':
                self.reserved_share_extend_percentage,
            'max_over_subscription_ratio':
                self.max_over_subscription_ratio
        }
        stats_dict['pools'].append(pool_stat)

    def _get_nfs_location(self, file_interfaces, export_path):
        export_locations = []
        for interface in file_interfaces:
            export_locations.append(
                {'path': f"{interface}:/{export_path}"})
        return export_locations

    def get_default_filter_function(self):
        return 'share.size >= 3'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add is_soft_deleted and scheduled_to_be_deleted_at to shares table

Revision ID: 1946cb97bb8d
Revises: fbdfabcba377
Create Date: 2021-07-14 14:41:58.615439

"""

# revision identifiers, used by Alembic.
revision = '1946cb97bb8d'
down_revision = 'fbdfabcba377'

from alembic import op
from oslo_log import log
import sqlalchemy as sa


LOG = log.getLogger(__name__)


def upgrade():
    try:
        op.add_column('shares', sa.Column(
            'is_soft_deleted', sa.Boolean,
            nullable=False, server_default=sa.sql.false()))
        op.add_column('shares', sa.Column(
            'scheduled_to_be_deleted_at', sa.DateTime))
    except Exception:
        LOG.error("Columns shares.is_soft_deleted "
                  "and/or shares.scheduled_to_be_deleted_at not created!")
        raise


def downgrade():
    try:
        op.drop_column('shares', 'is_soft_deleted')
        op.drop_column('shares', 'scheduled_to_be_deleted_at')
        LOG.warning("All shares in recycle bin will automatically be "
                    "restored, need to be manually identified and deleted "
                    "again.")
    except Exception:
        LOG.error("Column shares.is_soft_deleted and/or "
                  "shares.scheduled_to_be_deleted_at not dropped!")
        raise



# Copyright (c) 2022 China Telecom Digital Intelligence.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The share transfer api."""

from http import client as http_client

from oslo_log import log as logging
from oslo_utils import strutils
from oslo_utils import uuidutils
import webob
from webob import exc

from manila.api import common
from manila.api.openstack import wsgi
from manila.api.views import transfers as transfer_view
from manila import exception
from manila.i18n import _
from manila.transfer import api as transfer_api

LOG = logging.getLogger(__name__)
SHARE_TRANSFER_VERSION = "2.77"


class ShareTransferController(wsgi.Controller):
    """The Share Transfer API controller for the OpenStack API."""

    resource_name = 'share_transfer'
    _view_builder_class = transfer_view.ViewBuilder

    def __init__(self):
        self.transfer_api = transfer_api.API()
        super(ShareTransferController, self).__init__()

    @wsgi.Controller.authorize('get')
    @wsgi.Controller.api_version(SHARE_TRANSFER_VERSION)
    def show(self, req, id):
        """Return data about active transfers."""
        context = req.environ['manila.context']

        # Not found exception will be handled at the wsgi level
        transfer = self.transfer_api.get(context, transfer_id=id)

        return self._view_builder.detail(req, transfer)

    @wsgi.Controller.api_version(SHARE_TRANSFER_VERSION)
    def index(self, req):
        """Returns a summary list of transfers."""
        return self._get_transfers(req, is_detail=False)

    @wsgi.Controller.api_version(SHARE_TRANSFER_VERSION)
    def detail(self, req):
        """Returns a detailed list of transfers."""
        return self._get_transfers(req, is_detail=True)

    @wsgi.Controller.authorize('get_all')
    def _get_transfers(self, req, is_detail):
        """Returns a list of transfers, transformed through view builder."""
        context = req.environ['manila.context']
        params = req.params.copy()
        pagination_params = common.get_pagination_params(req)
        limit, offset = [pagination_params.pop('limit', None),
                         pagination_params.pop('offset', None)]
        sort_key, sort_dir = common.get_sort_params(params)

        filters = params
        key_map = {'name': 'display_name', 'name~': 'display_name~'}
        for k in key_map:
            if k in filters:
                filters[key_map[k]] = filters.pop(k)
        LOG.debug('Listing share transfers.')

        transfers = self.transfer_api.get_all(context,
                                              limit=limit,
                                              sort_key=sort_key,
                                              sort_dir=sort_dir,
                                              filters=filters,
                                              offset=offset)

        if is_detail:
            transfers = self._view_builder.detail_list(req, transfers)
        else:
            transfers = self._view_builder.summary_list(req, transfers)

        return transfers

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.Controller.api_version(SHARE_TRANSFER_VERSION)
    @wsgi.Controller.authorize('create')
    def create(self, req, body):
        """Create a new share transfer."""
        LOG.debug('Creating new share transfer %s', body)
        context = req.environ['manila.context']

        if not self.is_valid_body(body, 'transfer'):
            msg = _("'transfer' is missing from the request body.")
            raise exc.HTTPBadRequest(explanation=msg)

        transfer = body.get('transfer', {})

        share_id = transfer.get('share_id')
        if not share_id:
            msg = _("Must supply 'share_id' attribute.")
            raise exc.HTTPBadRequest(explanation=msg)
        if not uuidutils.is_uuid_like(share_id):
            msg = _("The 'share_id' attribute must be a uuid.")
            raise exc.HTTPBadRequest(explanation=msg)

        transfer_name = transfer.get('name')
        if transfer_name is not None:
            transfer_name = transfer_name.strip()

        LOG.debug("Creating transfer of share %s", share_id)

        try:
            new_transfer = self.transfer_api.create(context, share_id,
                                                    transfer_name)
        except exception.Invalid as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        transfer = self._view_builder.create(req,
                                             dict(new_transfer))
        return transfer

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.Controller.api_version(SHARE_TRANSFER_VERSION)
    @wsgi.Controller.authorize('accept')
    def accept(self, req, id, body):
        """Accept a new share transfer."""
        transfer_id = id
        LOG.debug('Accepting share transfer %s', transfer_id)
        context = req.environ['manila.context']

        if not self.is_valid_body(body, 'accept'):
            msg = _("'accept' is missing from the request body.")
            raise exc.HTTPBadRequest(explanation=msg)

        accept = body.get('accept', {})
        auth_key = accept.get('auth_key')
        if not auth_key:
            msg = _("Must supply 'auth_key' while accepting a "
                    "share transfer.")
            raise exc.HTTPBadRequest(explanation=msg)

        clear_rules = accept.get('clear_access_rules', False)
        if clear_rules:
            try:
                clear_rules = strutils.bool_from_string(clear_rules,
                                                        strict=True)
            except (ValueError, TypeError):
                msg = (_('Invalid boolean clear_access_rules : %(value)s') %
                       {'value': accept['clear_access_rules']})
                raise exc.HTTPBadRequest(explanation=msg)

        LOG.debug("Accepting transfer %s", transfer_id)

        try:
            self.transfer_api.accept(
                context, transfer_id, auth_key, clear_rules=clear_rules)
        except (exception.ShareSizeExceedsLimit,
                exception.ShareLimitExceeded,
                exception.ShareSizeExceedsAvailableQuota,
                exception.ShareReplicasLimitExceeded,
                exception.ShareReplicaSizeExceedsAvailableQuota,
                exception.SnapshotSizeExceedsAvailableQuota,
                exception.SnapshotLimitExceeded) as e:
            raise exc.HTTPRequestEntityTooLarge(explanation=e.msg,
                                                headers={'Retry-After': '0'})
        except (exception.InvalidShare,
                exception.InvalidSnapshot,
                exception.InvalidAuthKey,
                exception.TransferNotFound) as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

    @wsgi.Controller.api_version(SHARE_TRANSFER_VERSION)
    @wsgi.Controller.authorize('delete')
    def delete(self, req, id):
        """Delete a transfer."""
        context = req.environ['manila.context']

        LOG.debug("Delete transfer with id: %s", id)

        # Not found exception will be handled at the wsgi level
        self.transfer_api.delete(context, transfer_id=id)
        return webob.Response(status_int=http_client.OK)


def create_resource():
    return wsgi.Resource(ShareTransferController())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_backup_type_column

Revision ID: 2f27d904214c
Revises: 6e32091979e0
Create Date: 2024-03-10 23:16:18.130654

"""

# revision identifiers, used by Alembic.
revision = '2f27d904214c'
down_revision = '6e32091979e0'

from alembic import op
from oslo_log import log
import sqlalchemy as sa


LOG = log.getLogger(__name__)
share_backups_table_name = 'share_backups'
column_name = "backup_type"


def upgrade():
    try:
        op.add_column(share_backups_table_name,
                      sa.Column(column_name, sa.String(32), nullable=True))
    except Exception:
        LOG.error("Column 'backup_type' not created!")
        raise


def downgrade():
    try:
        op.drop_column(share_backups_table_name, column_name)
    except Exception:
        LOG.error("Column backup_type not dropped!")
        raise



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add backup

Revision ID: 9afbe2df4945
Revises: aebe2a413e13
Create Date: 2022-04-21 23:06:59.144695

"""

# revision identifiers, used by Alembic.
revision = '9afbe2df4945'
down_revision = 'aebe2a413e13'

from alembic import op
from oslo_log import log
import sqlalchemy as sa


LOG = log.getLogger(__name__)

share_backups_table_name = 'share_backups'


def upgrade():
    """Add backup attributes."""

    try:
        op.create_table(
            share_backups_table_name,
            sa.Column('id', sa.String(length=36),
                      primary_key=True, nullable=False),
            sa.Column('created_at', sa.DateTime),
            sa.Column('updated_at', sa.DateTime),
            sa.Column('deleted_at', sa.DateTime),
            sa.Column('deleted', sa.String(length=36), default='False'),
            sa.Column('user_id', sa.String(255)),
            sa.Column('project_id', sa.String(255)),
            sa.Column('availability_zone', sa.String(255)),
            sa.Column('fail_reason', sa.String(255)),
            sa.Column('display_name', sa.String(255)),
            sa.Column('display_description', sa.String(255)),
            sa.Column('host', sa.String(255)),
            sa.Column('topic', sa.String(255)),
            sa.Column('status', sa.String(255)),
            sa.Column('progress', sa.String(32)),
            sa.Column('restore_progress', sa.String(32)),
            sa.Column('size', sa.Integer),
            sa.Column('share_id', sa.String(36),
                      sa.ForeignKey('shares.id',
                                    name="fk_backups_share_id_shares")),
            mysql_engine='InnoDB',
            mysql_charset='utf8'
        )
    except Exception:
        LOG.error("Table |%s| not created!",
                  share_backups_table_name)
        raise

    try:
        op.add_column(
            'shares',
            sa.Column('source_backup_id', sa.String(36), nullable=True))
    except Exception:
        LOG.error("Column can not be added for 'shares' table!")
        raise


def downgrade():
    """Remove share backup attributes and table share_backups."""
    try:
        op.drop_table(share_backups_table_name)
    except Exception:
        LOG.error("%s table not dropped.", share_backups_table_name)
        raise

    try:
        op.drop_column('shares', 'source_backup_id')
    except Exception:
        LOG.error("Column can not be dropped for 'shares' table!")
        raise



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add default_ad_site to security service

Revision ID: c476aeb186ec
Revises: bb5938d74b73
Create Date: 2022-11-30 10:59:34.866946

"""

# revision identifiers, used by Alembic.
revision = 'c476aeb186ec'
down_revision = 'bb5938d74b73'

from alembic import op
from oslo_log import log
import sqlalchemy as sa

LOG = log.getLogger(__name__)

ss_table_name = 'security_services'


def upgrade():
    try:
        op.add_column(
            ss_table_name,
            sa.Column('default_ad_site', sa.String(255), nullable=True))
    except Exception:
        LOG.error("%s table column default_ad_site not added", ss_table_name)
        raise


def downgrade():
    try:
        op.drop_column(ss_table_name, 'default_ad_site')
    except Exception:
        LOG.error("%s table column default_ad_site not dropped", ss_table_name)
        raise



# Copyright 2015, Hitachi Data Systems.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests For Data Manager
"""

from unittest import mock

import ddt
from oslo_config import cfg

from manila.common import constants
from manila import context
from manila.data import helper
from manila.data import manager
from manila.data import utils as data_utils
from manila import db
from manila import exception
from manila import quota
from manila.share import rpcapi as share_rpc
from manila import test
from manila.tests import db_utils
from manila import utils


CONF = cfg.CONF


@ddt.ddt
class DataManagerTestCase(test.TestCase):
    """Test case for data manager."""

    def setUp(self):
        super(DataManagerTestCase, self).setUp()
        self.manager = manager.DataManager()
        self.context = context.get_admin_context()
        self.topic = 'fake_topic'
        self.share = db_utils.create_share()
        manager.CONF.set_default('mount_tmp_location', '/tmp/')
        manager.CONF.set_default('backup_mount_tmp_location', '/tmp/')
        manager.CONF.set_default(
            'backup_driver',
            'manila.tests.fake_backup_driver.FakeBackupDriver')

    def test_init(self):
        manager = self.manager
        self.assertIsNotNone(manager)

    @ddt.data(constants.TASK_STATE_DATA_COPYING_COMPLETING,
              constants.TASK_STATE_DATA_COPYING_STARTING,
              constants.TASK_STATE_DATA_COPYING_IN_PROGRESS)
    def test_init_host(self, status):

        share = db_utils.create_share(
            task_state=status)

        # mocks
        self.mock_object(db, 'share_get_all', mock.Mock(
            return_value=[share]))
        self.mock_object(db, 'share_update')

        # run
        self.manager.init_host()

        # asserts
        db.share_get_all.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext))

        db.share_update.assert_called_with(
            utils.IsAMatcher(context.RequestContext), share['id'],
            {'task_state': constants.TASK_STATE_DATA_COPYING_ERROR})

    @ddt.data(None, Exception('fake'), exception.ShareDataCopyCancelled())
    def test_migration_start(self, exc):

        migration_info_src = {
            'mount': 'mount_cmd_src',
            'unmount': 'unmount_cmd_src',
        }
        migration_info_dest = {
            'mount': 'mount_cmd_dest',
            'unmount': 'unmount_cmd_dest',
        }
        # mocks
        self.mock_object(db, 'share_get', mock.Mock(return_value=self.share))
        self.mock_object(db, 'share_instance_get', mock.Mock(
            return_value=self.share.instance))

        self.mock_object(data_utils, 'Copy',
                         mock.Mock(return_value='fake_copy'))

        if exc is None:
            self.manager.busy_tasks_shares[self.share['id']] = 'fake_copy'

        self.mock_object(self.manager, '_copy_share_data',
                         mock.Mock(side_effect=exc))

        self.mock_object(share_rpc.ShareAPI, 'migration_complete')

        if exc is not None and not isinstance(
                exc, exception.ShareDataCopyCancelled):
            self.mock_object(db, 'share_update')

        # run
        if exc is None or isinstance(exc, exception.ShareDataCopyCancelled):
            self.manager.migration_start(
                self.context, [], self.share['id'],
                'ins1_id', 'ins2_id', migration_info_src,
                migration_info_dest)
        else:
            self.assertRaises(
                exception.ShareDataCopyFailed, self.manager.migration_start,
                self.context, [], self.share['id'], 'ins1_id', 'ins2_id',
                migration_info_src, migration_info_dest)

            db.share_update.assert_called_once_with(
                self.context, self.share['id'],
                {'task_state': constants.TASK_STATE_DATA_COPYING_ERROR})

        # asserts
        self.assertFalse(self.manager.busy_tasks_shares.get(self.share['id']))

        if exc:
            share_rpc.ShareAPI.migration_complete.assert_called_once_with(
                self.context, self.share.instance, 'ins2_id')

    @ddt.data(
        {'cancelled': False, 'exc': None, 'case': 'migration'},
        {'cancelled': False, 'exc': Exception('fake'), 'case': 'migration'},
        {'cancelled': True, 'exc': None, 'case': 'migration'},
        {'cancelled': False, 'exc': None, 'case': 'backup'},
        {'cancelled': False, 'exc': Exception('fake'), 'case': 'backup'},
        {'cancelled': True, 'exc': None, 'case': 'backup'},
        {'cancelled': False, 'exc': None, 'case': 'restore'},
        {'cancelled': False, 'exc': Exception('fake'), 'case': 'restore'},
        {'cancelled': True, 'exc': None, 'case': 'restore'},
    )
    @ddt.unpack
    def test__copy_share_data(self, cancelled, exc, case):

        access = db_utils.create_access(share_id=self.share['id'])

        if case == 'migration':
            connection_info_src = {
                'mount': 'mount_cmd_src',
                'unmount': 'unmount_cmd_src',
                'share_id': self.share['id'],
                'share_instance_id': 'ins1_id',
                'mount_point': '/tmp/ins1_id',
            }
            connection_info_dest = {
                'mount': 'mount_cmd_dest',
                'unmount': 'unmount_cmd_dest',
                'share_id': None,
                'share_instance_id': 'ins2_id',
                'mount_point': '/tmp/ins2_id',
            }
        if case == 'backup':
            connection_info_src = {
                'mount': 'mount_cmd_src',
                'unmount': 'unmount_cmd_src',
                'share_id': self.share['id'],
                'share_instance_id': 'ins1_id',
                'mount_point': '/tmp/ins1_id',
            }
            connection_info_dest = {
                'mount': 'mount_cmd_dest',
                'unmount': 'unmount_cmd_dest',
                'share_id': None,
                'share_instance_id': None,
                'mount_point': '/tmp/backup_id',
                'backup': True
            }
        if case == 'restore':
            connection_info_src = {
                'mount': 'mount_cmd_src',
                'unmount': 'unmount_cmd_src',
                'share_id': None,
                'share_instance_id': None,
                'mount_point': '/tmp/backup_id',
                'restore': True
            }
            connection_info_dest = {
                'mount': 'mount_cmd_dest',
                'unmount': 'unmount_cmd_dest',
                'share_id': self.share['id'],
                'share_instance_id': 'ins2_id',
                'mount_point': '/tmp/ins2_id',
            }

        get_progress = {'total_progress': 100}

        # mocks
        fake_copy = mock.MagicMock(cancelled=cancelled)

        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_instance_get',
                         mock.Mock(side_effect=[self.share['instance'],
                                                self.share['instance']]))
        self.mock_object(helper.DataServiceHelper,
                         'allow_access_to_data_service',
                         mock.Mock(return_value=[access]))

        self.mock_object(helper.DataServiceHelper,
                         'mount_share_instance_or_backup')

        self.mock_object(fake_copy, 'run', mock.Mock(side_effect=exc))

        self.mock_object(fake_copy, 'get_progress',
                         mock.Mock(return_value=get_progress))

        self.mock_object(helper.DataServiceHelper,
                         'unmount_share_instance_or_backup',
                         mock.Mock(side_effect=Exception('fake')))

        self.mock_object(helper.DataServiceHelper,
                         'deny_access_to_data_service',
                         mock.Mock(side_effect=Exception('fake')))

        extra_updates = None

        # run
        if cancelled:
            self.assertRaises(
                exception.ShareDataCopyCancelled,
                self.manager._copy_share_data, self.context, fake_copy,
                connection_info_src, connection_info_dest)
            extra_updates = [
                mock.call(
                    self.context, self.share['id'],
                    {'task_state':
                     constants.TASK_STATE_DATA_COPYING_COMPLETING}),
                mock.call(
                    self.context, self.share['id'],
                    {'task_state':
                     constants.TASK_STATE_DATA_COPYING_CANCELLED})
            ]

        elif exc:
            self.assertRaises(
                exception.ShareDataCopyFailed, self.manager._copy_share_data,
                self.context, fake_copy, connection_info_src,
                connection_info_dest)

        else:
            self.manager._copy_share_data(
                self.context, fake_copy, connection_info_src,
                connection_info_dest)
            extra_updates = [
                mock.call(
                    self.context, self.share['id'],
                    {'task_state':
                     constants.TASK_STATE_DATA_COPYING_COMPLETING}),
                mock.call(
                    self.context, self.share['id'],
                    {'task_state':
                     constants.TASK_STATE_DATA_COPYING_COMPLETED})
            ]

        # asserts
        self.assertEqual(
            self.manager.busy_tasks_shares[self.share['id']], fake_copy)

        update_list = [
            mock.call(
                self.context, self.share['id'],
                {'task_state': constants.TASK_STATE_DATA_COPYING_STARTING}),
            mock.call(
                self.context, self.share['id'],
                {'task_state': constants.TASK_STATE_DATA_COPYING_IN_PROGRESS}),
        ]
        if extra_updates:
            update_list = update_list + extra_updates

        db.share_update.assert_has_calls(update_list)

        helper.DataServiceHelper.\
            mount_share_instance_or_backup.assert_has_calls([
                mock.call(connection_info_src, '/tmp/'),
                mock.call(connection_info_dest, '/tmp/')])

        fake_copy.run.assert_called_once_with()
        if exc is None:
            fake_copy.get_progress.assert_called_once_with()

        helper.DataServiceHelper.\
            unmount_share_instance_or_backup.assert_has_calls([
                mock.call(connection_info_src, '/tmp/'),
                mock.call(connection_info_dest, '/tmp/')])

    def test__copy_share_data_exception_access(self):

        connection_info_src = {
            'mount': 'mount_cmd_src',
            'unmount': 'unmount_cmd_src',
            'share_id': self.share['id'],
            'share_instance_id': 'ins1_id',
            'mount_point': '/tmp/ins1_id',
        }
        connection_info_dest = {
            'mount': 'mount_cmd_dest',
            'unmount': 'unmount_cmd_dest',
            'share_id': None,
            'share_instance_id': 'ins2_id',
            'mount_point': '/tmp/ins2_id',
        }

        fake_copy = mock.MagicMock(cancelled=False)

        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_instance_get',
                         mock.Mock(side_effect=[self.share['instance'],
                                                self.share['instance']]))

        self.mock_object(
            helper.DataServiceHelper, 'allow_access_to_data_service',
            mock.Mock(
                side_effect=exception.ShareDataCopyFailed(reason='fake')))

        self.mock_object(helper.DataServiceHelper, 'cleanup_data_access')

        # run
        self.assertRaises(exception.ShareDataCopyFailed,
                          self.manager._copy_share_data, self.context,
                          fake_copy, connection_info_src, connection_info_dest)

        # asserts
        db.share_update.assert_called_once_with(
            self.context, self.share['id'],
            {'task_state': constants.TASK_STATE_DATA_COPYING_STARTING})

        (helper.DataServiceHelper.allow_access_to_data_service.
            assert_called_once_with(
                self.share['instance'], connection_info_src,
                self.share['instance'], connection_info_dest))

    def test__copy_share_data_exception_mount_1(self):

        access = db_utils.create_access(share_id=self.share['id'])

        connection_info_src = {
            'mount': 'mount_cmd_src',
            'unmount': 'unmount_cmd_src',
            'share_id': self.share['id'],
            'share_instance_id': 'ins1_id',
            'mount_point': '/tmp/ins1_id',
        }
        connection_info_dest = {
            'mount': 'mount_cmd_dest',
            'unmount': 'unmount_cmd_dest',
            'share_id': None,
            'share_instance_id': 'ins2_id',
            'mount_point': '/tmp/ins2_id',
        }

        fake_copy = mock.MagicMock(cancelled=False)

        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_instance_get',
                         mock.Mock(side_effect=[self.share['instance'],
                                                self.share['instance']]))

        self.mock_object(helper.DataServiceHelper,
                         'allow_access_to_data_service',
                         mock.Mock(return_value=[access]))

        self.mock_object(helper.DataServiceHelper,
                         'mount_share_instance_or_backup',
                         mock.Mock(side_effect=Exception('fake')))

        self.mock_object(helper.DataServiceHelper, 'cleanup_data_access')
        self.mock_object(helper.DataServiceHelper, 'cleanup_temp_folder')

        # run
        self.assertRaises(exception.ShareDataCopyFailed,
                          self.manager._copy_share_data, self.context,
                          fake_copy, connection_info_src, connection_info_dest)

        # asserts
        db.share_update.assert_called_once_with(
            self.context, self.share['id'],
            {'task_state': constants.TASK_STATE_DATA_COPYING_STARTING})

        helper.DataServiceHelper.\
            mount_share_instance_or_backup.assert_called_once_with(
                connection_info_src, '/tmp/')

        helper.DataServiceHelper.cleanup_temp_folder.assert_called_once_with(
            '/tmp/', 'ins1_id')

        helper.DataServiceHelper.cleanup_data_access.assert_has_calls([
            mock.call([access], self.share['instance']),
            mock.call([access], self.share['instance'])])

    def test__copy_share_data_exception_mount_2(self):

        access = db_utils.create_access(share_id=self.share['id'])

        connection_info_src = {
            'mount': 'mount_cmd_src',
            'unmount': 'unmount_cmd_src',
            'share_id': self.share['id'],
            'share_instance_id': 'ins1_id',
            'mount_point': '/tmp/ins1_id',
        }
        connection_info_dest = {
            'mount': 'mount_cmd_dest',
            'unmount': 'unmount_cmd_dest',
            'share_id': None,
            'share_instance_id': 'ins2_id',
            'mount_point': '/tmp/ins2_id',
        }

        fake_copy = mock.MagicMock(cancelled=False)

        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_instance_get',
                         mock.Mock(side_effect=[self.share['instance'],
                                                self.share['instance']]))

        self.mock_object(helper.DataServiceHelper,
                         'allow_access_to_data_service',
                         mock.Mock(return_value=[access]))

        self.mock_object(helper.DataServiceHelper,
                         'mount_share_instance_or_backup',
                         mock.Mock(side_effect=[None, Exception('fake')]))

        self.mock_object(helper.DataServiceHelper, 'cleanup_data_access')
        self.mock_object(helper.DataServiceHelper, 'cleanup_temp_folder')
        self.mock_object(helper.DataServiceHelper,
                         'cleanup_unmount_temp_folder')

        # run
        self.assertRaises(exception.ShareDataCopyFailed,
                          self.manager._copy_share_data, self.context,
                          fake_copy, connection_info_src, connection_info_dest)

        # asserts
        db.share_update.assert_called_once_with(
            self.context, self.share['id'],
            {'task_state': constants.TASK_STATE_DATA_COPYING_STARTING})

        helper.DataServiceHelper.\
            mount_share_instance_or_backup.assert_has_calls([
                mock.call(connection_info_src, '/tmp/'),
                mock.call(connection_info_dest, '/tmp/')])

        helper.DataServiceHelper.cleanup_unmount_temp_folder.\
            assert_called_once_with(connection_info_src, '/tmp/')

        helper.DataServiceHelper.cleanup_temp_folder.assert_has_calls([
            mock.call('/tmp/', 'ins2_id'), mock.call('/tmp/', 'ins1_id')])

    def test_data_copy_cancel(self):

        share = db_utils.create_share()

        self.manager.busy_tasks_shares[share['id']] = data_utils.Copy

        # mocks
        self.mock_object(data_utils.Copy, 'cancel')

        # run
        self.manager.data_copy_cancel(self.context, share['id'])

        # asserts
        data_utils.Copy.cancel.assert_called_once_with()

    def test_data_copy_cancel_not_copying(self):

        self.assertRaises(exception.InvalidShare,
                          self.manager.data_copy_cancel, self.context,
                          'fake_id')

    def test_data_copy_get_progress(self):

        share = db_utils.create_share()

        self.manager.busy_tasks_shares[share['id']] = data_utils.Copy

        expected = 'fake_progress'

        # mocks
        self.mock_object(data_utils.Copy, 'get_progress',
                         mock.Mock(return_value=expected))

        # run
        result = self.manager.data_copy_get_progress(self.context, share['id'])

        # asserts
        self.assertEqual(expected, result)

        data_utils.Copy.get_progress.assert_called_once_with()

    def test_data_copy_get_progress_not_copying(self):

        self.assertRaises(exception.InvalidShare,
                          self.manager.data_copy_get_progress, self.context,
                          'fake_id')

    def test_create_share_backup(self):
        share_info = db_utils.create_share(
            status=constants.STATUS_BACKUP_CREATING)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_CREATING)

        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_backup_update')
        self.mock_object(db, 'share_get', mock.Mock(return_value=share_info))
        self.mock_object(db, 'share_backup_get',
                         mock.Mock(return_value=backup_info))
        self.mock_object(self.manager, '_run_backup',
                         mock.Mock(side_effect=None))
        self.manager.create_backup(self.context, backup_info)
        db.share_update.assert_called_with(
            self.context, share_info['id'],
            {'status': constants.STATUS_AVAILABLE})
        db.share_backup_update.assert_called_with(
            self.context, backup_info['id'],
            {'status': constants.STATUS_AVAILABLE, 'progress': '100'})

    def test_create_share_backup_exception(self):
        share_info = db_utils.create_share(status=constants.STATUS_AVAILABLE)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_AVAILABLE, size=2)

        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_backup_update')
        self.mock_object(db, 'share_get', mock.Mock(return_value=share_info))
        self.mock_object(db, 'share_backup_get',
                         mock.Mock(return_value=backup_info))
        self.mock_object(
            self.manager, '_run_backup',
            mock.Mock(
                side_effect=exception.ShareDataCopyFailed(reason='fake')))
        self.assertRaises(exception.ManilaException,
                          self.manager.create_backup,
                          self.context, backup_info)
        db.share_update.assert_called_with(
            self.context, share_info['id'],
            {'status': constants.STATUS_AVAILABLE})
        db.share_backup_update.assert_called()

    @ddt.data('90', '100')
    def test_create_share_backup_continue(self, progress):
        share_info = db_utils.create_share(
            status=constants.STATUS_BACKUP_CREATING)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_CREATING,
            topic=CONF.data_topic)
        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_backup_update')
        self.mock_object(db, 'share_backups_get_all',
                         mock.Mock(return_value=[backup_info]))
        self.mock_object(self.manager, 'data_copy_get_progress',
                         mock.Mock(return_value={'total_progress': progress}))

        self.manager.create_backup_continue(self.context)
        if progress == '100':
            db.share_backup_update.assert_called_with(
                self.context, backup_info['id'],
                {'status': constants.STATUS_AVAILABLE, 'progress': '100'})
            db.share_update.assert_called_with(
                self.context, share_info['id'],
                {'status': constants.STATUS_AVAILABLE})
        else:
            db.share_backup_update.assert_called_with(
                self.context, backup_info['id'],
                {'progress': progress})

    def test_create_share_backup_continue_exception(self):
        share_info = db_utils.create_share(
            status=constants.STATUS_BACKUP_CREATING)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_CREATING,
            topic=CONF.data_topic)
        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_backup_update')
        self.mock_object(db, 'share_backups_get_all',
                         mock.Mock(return_value=[backup_info]))
        self.mock_object(self.manager, 'data_copy_get_progress',
                         mock.Mock(side_effect=exception.ManilaException))

        self.manager.create_backup_continue(self.context)

        db.share_backup_update.assert_called_with(
            self.context, backup_info['id'],
            {'status': constants.STATUS_ERROR, 'progress': '0'})
        db.share_update.assert_called_with(
            self.context, share_info['id'],
            {'status': constants.STATUS_AVAILABLE})

    @ddt.data(None, exception.ShareDataCopyFailed(reason='fake'))
    def test__run_backup(self, exc):
        share_info = db_utils.create_share(status=constants.STATUS_AVAILABLE)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_AVAILABLE, size=2)
        share_instance = {
            'export_locations': [{
                'path': 'test_path',
                "is_admin_only": False
                }, ],
            'share_proto': 'nfs',
        }

        # mocks
        self.mock_object(db, 'share_instance_get',
                         mock.Mock(return_value=share_instance))

        self.mock_object(data_utils, 'Copy',
                         mock.Mock(return_value='fake_copy'))
        self.manager.busy_tasks_shares[self.share['id']] = 'fake_copy'
        self.mock_object(self.manager, '_copy_share_data',
                         mock.Mock(side_effect=exc))

        self.mock_object(self.manager, '_run_backup')

        if exc is isinstance(exc, exception.ShareDataCopyFailed):
            self.assertRaises(exception.ShareDataCopyFailed,
                              self.manager._run_backup, self.context,
                              backup_info, share_info)
        else:
            self.manager._run_backup(self.context, backup_info, share_info)

    def test_delete_share_backup(self):
        share_info = db_utils.create_share(status=constants.STATUS_AVAILABLE)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_AVAILABLE, size=2)
        # mocks
        self.mock_object(db, 'share_backup_delete')
        self.mock_object(db, 'share_backup_get',
                         mock.Mock(return_value=backup_info))
        self.mock_object(utils, 'execute')

        reservation = 'fake'
        self.mock_object(quota.QUOTAS, 'reserve',
                         mock.Mock(return_value=reservation))
        self.mock_object(quota.QUOTAS, 'commit')

        self.manager.delete_backup(self.context, backup_info)
        db.share_backup_delete.assert_called_with(
            self.context, backup_info['id'])

    def test_delete_share_backup_exception(self):
        share_info = db_utils.create_share(status=constants.STATUS_AVAILABLE)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_AVAILABLE, size=2)
        # mocks
        self.mock_object(db, 'share_backup_get',
                         mock.Mock(return_value=backup_info))
        self.mock_object(utils, 'execute')

        self.mock_object(
            quota.QUOTAS, 'reserve',
            mock.Mock(side_effect=exception.ManilaException))
        self.assertRaises(exception.ManilaException,
                          self.manager.delete_backup, self.context,
                          backup_info)

    def test_restore_share_backup(self):
        share_info = db_utils.create_share(status=constants.STATUS_AVAILABLE)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_AVAILABLE, size=2)
        share_id = share_info['id']

        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_backup_update')
        self.mock_object(db, 'share_get', mock.Mock(return_value=share_info))
        self.mock_object(db, 'share_backup_get',
                         mock.Mock(return_value=backup_info))
        self.mock_object(self.manager, '_run_restore')
        self.manager.restore_backup(self.context, backup_info, share_id)
        db.share_update.assert_called_with(
            self.context, share_info['id'],
            {'status': constants.STATUS_AVAILABLE})
        db.share_backup_update.assert_called_with(
            self.context, backup_info['id'],
            {'status': constants.STATUS_AVAILABLE, 'restore_progress': '100'})

    def test_restore_share_backup_exception(self):
        share_info = db_utils.create_share(status=constants.STATUS_AVAILABLE)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_AVAILABLE, size=2)
        share_id = share_info['id']

        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_get', mock.Mock(return_value=share_info))
        self.mock_object(db, 'share_backup_get',
                         mock.Mock(return_value=backup_info))
        self.mock_object(
            self.manager, '_run_restore',
            mock.Mock(
                side_effect=exception.ShareDataCopyFailed(reason='fake')))
        self.assertRaises(exception.ManilaException,
                          self.manager.restore_backup, self.context,
                          backup_info, share_id)
        db.share_update.assert_called_with(
            self.context, share_info['id'],
            {'status': constants.STATUS_BACKUP_RESTORING_ERROR})

    @ddt.data('90', '100')
    def test_restore_share_backup_continue(self, progress):
        share_info = db_utils.create_share(
            status=constants.STATUS_BACKUP_RESTORING)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_RESTORING,
            topic=CONF.data_topic)
        share_info['source_backup_id'] = backup_info['id']

        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_backup_update')
        self.mock_object(db, 'share_get_all',
                         mock.Mock(return_value=[share_info]))
        self.mock_object(db, 'share_backups_get_all',
                         mock.Mock(return_value=[backup_info]))
        self.mock_object(self.manager, 'data_copy_get_progress',
                         mock.Mock(return_value={'total_progress': progress}))

        self.manager.restore_backup_continue(self.context)

        if progress == '100':
            db.share_backup_update.assert_called_with(
                self.context, backup_info['id'],
                {'status': constants.STATUS_AVAILABLE,
                 'restore_progress': '100'})
            db.share_update.assert_called_with(
                self.context, share_info['id'],
                {'status': constants.STATUS_AVAILABLE})
        else:
            db.share_backup_update.assert_called_with(
                self.context, backup_info['id'],
                {'restore_progress': progress})

    def test_restore_share_backup_continue_exception(self):
        share_info = db_utils.create_share(
            status=constants.STATUS_BACKUP_RESTORING)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_RESTORING,
            topic=CONF.data_topic)
        share_info['source_backup_id'] = backup_info['id']

        # mocks
        self.mock_object(db, 'share_update')
        self.mock_object(db, 'share_backup_update')
        self.mock_object(db, 'share_get_all',
                         mock.Mock(return_value=[share_info]))
        self.mock_object(db, 'share_backups_get_all',
                         mock.Mock(return_value=[backup_info]))
        self.mock_object(self.manager, 'data_copy_get_progress',
                         mock.Mock(side_effect=exception.ManilaException))

        self.manager.restore_backup_continue(self.context)
        db.share_backup_update.assert_called_with(
            self.context, backup_info['id'],
            {'status': constants.STATUS_AVAILABLE, 'restore_progress': '0'})
        db.share_update.assert_called_with(
            self.context, share_info['id'],
            {'status': constants.STATUS_BACKUP_RESTORING_ERROR})

    @ddt.data(None, exception.ShareDataCopyFailed(reason='fake'))
    def test__run_restore(self, exc):
        share_info = db_utils.create_share(status=constants.STATUS_AVAILABLE)
        backup_info = db_utils.create_backup(
            share_info['id'], status=constants.STATUS_AVAILABLE, size=2)
        share_instance = {
            'export_locations': [{
                'path': 'test_path',
                "is_admin_only": False
                }, ],
            'share_proto': 'nfs',
        }

        # mocks
        self.mock_object(db, 'share_instance_get',
                         mock.Mock(return_value=share_instance))

        self.mock_object(data_utils, 'Copy',
                         mock.Mock(return_value='fake_copy'))
        self.manager.busy_tasks_shares[self.share['id']] = 'fake_copy'
        self.mock_object(self.manager, '_copy_share_data',
                         mock.Mock(side_effect=exc))

        self.mock_object(self.manager, '_run_restore')

        if exc is isinstance(exc, exception.ShareDataCopyFailed):
            self.assertRaises(exception.ShareDataCopyFailed,
                              self.manager._run_restore, self.context,
                              backup_info, share_info)
        else:
            self.manager._run_restore(self.context, backup_info, share_info)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

import ddt
from oslo_config import cfg
from oslo_utils import uuidutils
import webob

from manila.api.v2 import resource_locks
from manila import context
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes
from manila.tests.api.v2 import stubs
from manila.tests import utils as test_utils
from manila import utils

CONF = cfg.CONF


@ddt.ddt
class ResourceLockApiTest(test.TestCase):
    def setUp(self):
        super(ResourceLockApiTest, self).setUp()
        self.controller = resource_locks.ResourceLocksController()
        self.maxDiff = None
        self.ctxt = context.RequestContext('demo', 'fake', False)
        self.req = fakes.HTTPRequest.blank(
            '/resource-locks',
            version=resource_locks.RESOURCE_LOCKS_MIN_API_VERSION
        )
        self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True)
        )

    @ddt.data(
        test_utils.annotated(
            'no_body_content', {
                'body': {},
                'resource_type': 'share'
            }
        ),
        test_utils.annotated(
            'invalid_body', {
                'body': {
                    'share': 'somedata'
                },
                'resource_type': 'share'
            }
        ),
        test_utils.annotated(
            'invalid_action', {
                'body': {
                    'resource_lock': {
                        'resource_action': 'invalid_action',
                    }
                },
                'resource_type': 'share'
            },
        ),
        test_utils.annotated(
            'invalid_reason', {
                'body': {
                    'resource_lock': {
                        'lock_reason': 'xyzzyspoon!' * 94,
                    }
                },
                'resource_type': 'share'
            },
        ),
        test_utils.annotated(
            'disallowed_attributes', {
                'body': {
                    'resource_lock': {
                        'lock_reason': 'the reason is you',
                        'resource_action': 'delete',
                        'resource_id': uuidutils.generate_uuid(),
                    }
                },
                'resource_type': 'share'
            },
        ),
    )
    @ddt.unpack
    def test__check_body_for_update_invalid(self, body, resource_type):
        current_lock = {'resource_type': resource_type}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._check_body,
                          body,
                          lock_to_update=current_lock)

    @ddt.data(
        test_utils.annotated('no_body_content', {}),
        test_utils.annotated('invalid_body', {'share': 'somedata'}),
        test_utils.annotated(
            'invalid_action', {
                'resource_lock': {
                    'resource_action': 'invalid_action',
                },
            },
        ),
        test_utils.annotated(
            'invalid_reason', {
                'resource_lock': {
                    'lock_reason': 'xyzzyspoon!' * 94,
                },
            },
        ),
        test_utils.annotated(
            'invalid_resource_id', {
                'resource_lock': {
                    'resource_id': 'invalid-id',
                    'resource_action': 'delete',
                },
            },
        ),
        test_utils.annotated(
            'invalid_resource_type', {
                'resource_lock': {
                    'resource_id': uuidutils.generate_uuid(),
                    'resource_type': 'invalid-resource-type',
                },
            },
        ),
    )
    def test__check_body_for_create_invalid(self, body):
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller._check_body,
                          body)

    @ddt.data(
        test_utils.annotated(
            'action_and_lock_reason', {
                'body': {
                    'resource_lock': {
                        'resource_action': 'delete',
                        'lock_reason': 'the reason is you',
                    }
                },
                'resource_type': 'share',
            },
        ),
        test_utils.annotated(
            'lock_reason', {
                'body': {
                    'resource_lock': {
                        'lock_reason': 'tienes razon',
                    }
                },
                'resource_type': 'share',
            },
        ),
        test_utils.annotated(
            'resource_action', {
                'body': {
                    'resource_lock': {
                        'resource_action': 'delete',
                    }
                },
                'resource_type': 'access_rule',
            },
        ),
    )
    @ddt.unpack
    def test__check_body_for_update(self, body, resource_type):
        current_lock = copy.copy(body['resource_lock'])
        current_lock['resource_type'] = resource_type

        result = self.controller._check_body(
            body, lock_to_update=current_lock)

        self.assertIsNone(result)

    def test__check_body_for_create(self):
        body = {
            'resource_lock': {
                'resource_id': uuidutils.generate_uuid(),
                'resource_type': 'share',
            },
        }

        result = self.controller._check_body(body)

        self.assertIsNone(result)

    @ddt.data({'created_since': None, 'created_before': None},
              {'created_since': '2222-22-22', 'created_before': 'a_year_ago'},
              {'created_since': 'epoch'},
              {'created_before': 'december'})
    def test_index_invalid_time_filters(self, filters):
        url = '/resource-locks?'
        for key, value in filters.items():
            url += f'{key}={value}&'
        url.rstrip('&')
        req = fakes.HTTPRequest.blank(
            url, version=resource_locks.RESOURCE_LOCKS_MIN_API_VERSION)
        req.environ['manila.context'] = self.ctxt

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.index,
                          req)

    @ddt.data({'limit': 'a', 'offset': 'test'},
              {'limit': -1},
              {'with_count': 'oh-noes', 'limit': 0})
    def test_index_invalid_pagination(self, filters):
        url = '/resource-locks?'
        for key, value in filters.items():
            url += f'{key}={value}&'
        url.rstrip('&')

        req = fakes.HTTPRequest.blank(
            url, version=resource_locks.RESOURCE_LOCKS_MIN_API_VERSION)
        req.environ['manila.context'] = self.ctxt

        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.index,
                          req)

    def test_index(self):
        url = ('/resource-locks?sort_dir=asc&sort_key=resource_id&limit=3'
               '&offset=1&project_id=f63f7a159f404cfc8604b7065c609691'
               '&with_count=1')
        req = fakes.HTTPRequest.blank(
            url, version=resource_locks.RESOURCE_LOCKS_MIN_API_VERSION)
        locks = [
            stubs.stub_lock('68e2e33d-0f0c-49b7-aee3-f0696ab90360'),
            stubs.stub_lock('93748a9f-6dfe-4baf-ad4c-b9c82d6063ef'),
            stubs.stub_lock('44f8dd68-2eeb-41df-b5d1-9e7654212527'),
        ]
        self.mock_object(self.controller.resource_locks_api,
                         'get_all',
                         mock.Mock(return_value=(locks, 3)))

        actual_locks = self.controller.index(req)

        expected_filters = {
            'project_id': 'f63f7a159f404cfc8604b7065c609691',
        }
        self.controller.resource_locks_api.get_all.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            search_opts=mock.ANY,
            limit=3,
            offset=1,
            sort_key='resource_id',
            sort_dir='asc',
            show_count=True,
        )
        # webob uses a "MultiDict" for request params
        actual_filters = {}
        call_args = self.controller.resource_locks_api.get_all.call_args[1]
        search_opts = call_args['search_opts']
        for key, value in search_opts.dict_of_lists().items():
            actual_filters[key] = value[0]

        self.assertEqual(expected_filters, actual_filters)
        self.assertEqual(3, len(actual_locks['resource_locks']))
        for lock in actual_locks['resource_locks']:
            for key in locks[0].keys():
                self.assertIn(key, lock)
            self.assertIn('links', lock)
        self.assertIn('resource_locks_links', actual_locks)
        self.assertEqual(3, actual_locks['count'])

    def test_show_not_found(self):
        url = '/resource-locks/fake-lock-id'
        req = fakes.HTTPRequest.blank(
            url, version=resource_locks.RESOURCE_LOCKS_MIN_API_VERSION)
        self.mock_object(
            self.controller.resource_locks_api, 'get',
            mock.Mock(side_effect=exception.ResourceLockNotFound(lock_id='1')))
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.show,
                          req,
                          'fake-lock-id')

    def test_show(self):
        url = '/resource-locks/c6aef27b-f583-48c7-aac1-bd8fb570ce16'
        req = fakes.HTTPRequest.blank(
            url, version=resource_locks.RESOURCE_LOCKS_MIN_API_VERSION)
        expected_lock = stubs.stub_lock(
            'c6aef27b-f583-48c7-aac1-bd8fb570ce16'
        )
        self.mock_object(
            self.controller.resource_locks_api,
            'get',
            mock.Mock(return_value=expected_lock)
        )

        actual_lock = self.controller.show(
            req, 'c6aef27b-f583-48c7-aac1-bd8fb570ce16')
        self.assertSubDictMatch(expected_lock, actual_lock['resource_lock'])
        self.assertIn('links', actual_lock['resource_lock'])

    def test_delete_not_found(self):
        url = '/resource-locks/fake-lock-id'
        req = fakes.HTTPRequest.blank(
            url, version=resource_locks.RESOURCE_LOCKS_MIN_API_VERSION)
        self.mock_object(
            self.controller.resource_locks_api,
            'delete',
            mock.Mock(side_effect=exception.ResourceLockNotFound(lock_id='1')),
        )
        self.assertRaises(webob.exc.HTTPNotFound,
                          self.controller.delete,
                          req,
                          'fake-lock-id')

    def test_delete(self):
        url = '/resource-locks/c6aef27b-f583-48c7-aac1-bd8fb570ce16'
        req = fakes.HTTPRequest.blank(
            url, version=resource_locks.RESOURCE_LOCKS_MIN_API_VERSION)
        self.mock_object(self.controller.resource_locks_api, 'delete')

        result = self.controller.delete(req,
                                        'c6aef27b-f583-48c7-aac1-bd8fb570ce16')
        self.assertEqual(204, result.status_int)

    def test_create_no_such_resource(self):
        self.mock_object(self.controller, '_check_body')
        body = {
            'resource_lock': {
                'resource_id': '27e14086-16e1-445b-ad32-b2ebb07225a8',
                'resource_type': 'share',
            },
        }
        self.mock_object(self.controller.resource_locks_api,
                         'create',
                         mock.Mock(side_effect=exception.NotFound))
        self.assertRaises(webob.exc.HTTPBadRequest,
                          self.controller.create,
                          self.req,
                          body)

    def test_create_visibility_already_locked(self):
        self.mock_object(self.controller, '_check_body')
        resource_id = '27e14086-16e1-445b-ad32-b2ebb07225a8'
        body = {
            'resource_lock': {
                'resource_id': resource_id,
                'resource_type': 'share',
            },
        }
        self.mock_object(
            self.controller.resource_locks_api,
            'create',
            mock.Mock(
                side_effect=exception.ResourceVisibilityLockExists(
                    resource_id=resource_id))
        )
        self.assertRaises(webob.exc.HTTPConflict,
                          self.controller.create,
                          self.req,
                          body)

    def test_create(self):
        self.mock_object(self.controller, '_check_body')
        expected_lock = stubs.stub_lock(
            '04512dae-18c2-45b5-bbab-50b775ba6f1d',
            lock_reason=None,
        )
        body = {
            'resource_lock': {
                'resource_id': expected_lock['resource_id'],
                'resource_type': expected_lock['resource_type'],
            },
        }
        self.mock_object(self.controller.resource_locks_api,
                         'create',
                         mock.Mock(return_value=expected_lock))

        actual_lock = self.controller.create(self.req, body)['resource_lock']

        self.controller.resource_locks_api.create.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            resource_id=expected_lock['resource_id'],
            resource_type=expected_lock['resource_type'],
            resource_action='delete',
            lock_reason=None,
        )
        self.assertSubDictMatch(expected_lock, actual_lock)
        self.assertIn('links', actual_lock)

    def test_update(self):
        expected_lock = stubs.stub_lock(
            '04512dae-18c2-45b5-bbab-50b775ba6f1d',
            lock_reason=None,
        )
        self.mock_object(self.controller, '_check_body')
        self.mock_object(self.controller.resource_locks_api, 'get',
                         mock.Mock(return_value=expected_lock))
        self.mock_object(self.controller.resource_locks_api,
                         'update',
                         mock.Mock(return_value=expected_lock))

        body = {
            'resource_lock': {
                'lock_reason': None
            },
        }

        actual_lock = self.controller.update(
            self.req,
            '04512dae-18c2-45b5-bbab-50b775ba6f1d',
            body
        )['resource_lock']

        self.controller.resource_locks_api.update.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext),
            expected_lock,
            {'lock_reason': None}
        )
        self.assertSubDictMatch(expected_lock, actual_lock)



# Copyright (c) 2022 MacroSAN Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Share driver for Macrosan Storage Array.
"""

import functools
from oslo_config import cfg
from oslo_log import log

from manila.share import driver
from manila.share.drivers.macrosan import macrosan_helper

macrosan_opts = [
    cfg.HostAddressOpt('macrosan_nas_ip',
                       required=True,
                       help='IP address for the Macrosan NAS server.'),
    cfg.PortOpt('macrosan_nas_port',
                default=8443,
                help='Port number for the Macrosan NAS server.'),
    cfg.StrOpt('macrosan_nas_username',
               default='manila',
               help='Username for the Macrosan NAS server.'),
    cfg.StrOpt('macrosan_nas_password',
               secret=True,
               help='Password for the Macrosan NAS server.'),
    cfg.StrOpt('macrosan_nas_http_protocol',
               default='https',
               choices=['http', 'https'],
               help='Http protocol for the Macrosan NAS server.'),
    cfg.BoolOpt('macrosan_ssl_cert_verify',
                default=False,
                help='Defines whether the driver should check ssl cert.'),
    cfg.StrOpt('macrosan_nas_prefix',
               default='nas',
               help='Url prefix for the Macrosan NAS server.'),
    cfg.ListOpt('macrosan_share_pools',
                required=True,
                help='Comma separated list of Macrosan NAS pools.'),
    cfg.IntOpt('macrosan_timeout',
               default=60,
               help='request timeout in seconds.')
]

CONF = cfg.CONF
CONF.register_opts(macrosan_opts)
LOG = log.getLogger(__name__)


def debug_trace(func):
    """Log the dirver invoke method start and leave information

    Used in the MacrosanNasDriver class methods.
    Ensure func have 'self' argument.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        driver = args[0]
        method_name = "%(class_name)s.%(method)s" \
                      % {"class_name": driver.__class__.__name__,
                         "method": func.__name__}
        backend_name = driver.configuration.share_backend_name
        LOG.debug("[%(backend_name)s]:Start %(method_name)s",
                  {"backend_name": backend_name, "method_name": method_name})
        result = func(*args, **kwargs)
        LOG.debug("[%(backend_name)s]:Leave %(method_name)s",
                  {"backend_name": backend_name, "method_name": method_name})
        return result

    return wrapper


class MacrosanNasDriver(driver.ShareDriver):
    """Macrosan Share Driver

        Driver version history:
        V1.0.0:    Initial version
                   Driver support:
                       share create/delete,
                       extend size,
                       shrink size,
                       update_access.
                       protocol: NFS/CIFS

    """

    VENDOR = 'Macrosan'
    VERSION = '1.0.0'
    PROTOCOL = 'NFS_CIFS'

    def __init__(self, *args, **kwargs):
        super(MacrosanNasDriver, self).__init__(False, *args, **kwargs)
        self.configuration.append_config_values(macrosan_opts)

        self.helper = macrosan_helper.MacrosanHelper(self.configuration)

    @debug_trace
    def do_setup(self, context):
        """initialization the driver when start"""
        self.helper.do_setup()

    @debug_trace
    def check_for_setup_error(self):
        """Check prerequisites"""
        self.helper.check_share_service()

    @debug_trace
    def create_share(self, context, share, share_server=None):
        """Create a share"""
        return self.helper.create_share(share, share_server)

    @debug_trace
    def delete_share(self, context, share, share_server=None):
        """Delete a share."""
        self.helper.delete_share(share, share_server)

    @debug_trace
    def extend_share(self, share, new_size, share_server=None):
        """Extend share capacity"""
        self.helper.extend_share(share, new_size, share_server)

    @debug_trace
    def shrink_share(self, share, new_size, share_server=None):
        """Shrink share capacity"""
        self.helper.shrink_share(share, new_size, share_server)

    @debug_trace
    def ensure_share(self, context, share, share_server=None):
        """Enusre that share is exported."""
        return self.helper.ensure_share(share, share_server)

    @debug_trace
    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules list.

        :param context: Current context
        :param share: Share model with share data.
        :param access_rules: All access rules for given share
        :param add_rules: Empty List or List of access rules which should be
               added. access_rules already contains these rules.
        :param delete_rules: Empty List or List of access rules which should be
               removed. access_rules doesn't contain these rules.
        :param share_server: Not used by this driver.

        :returns: None, or a dictionary of ``access_id``, ``access_key`` as
                  key: value pairs for the rules added, where, ``access_id``
                  is the UUID (string) of the access rule, and ``access_key``
                  is the credential (string) of the entity granted access.
                  During recovery after error, the returned dictionary must
                  contain ``access_id``, ``access_key`` for all the rules that
                  the driver is ordered to resync, i.e. rules in the
                  ``access_rules`` parameter.
        """
        return self.helper.update_access(share, access_rules,
                                         add_rules, delete_rules,
                                         share_server)

    @debug_trace
    def _update_share_stats(self):
        """Update backend status ,include driver and pools"""

        data = {
            'vendor_name': self.VENDOR,
            'driver_version': self.VERSION,
            'storage_protocol': self.PROTOCOL,
            'share_backend_name':
                self.configuration.safe_get('share_backend_name'),
        }
        self.helper.update_share_stats(data)
        super(MacrosanNasDriver, self)._update_share_stats(data)



# Copyright 2015, Hitachi Data Systems.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Data Service
"""

import os
import shutil

from oslo_config import cfg
from oslo_config import types
from oslo_log import log
from oslo_service import periodic_task
from oslo_utils import excutils
from oslo_utils import importutils

from manila.common import constants
from manila import context
from manila.data import helper
from manila.data import utils as data_utils
from manila import exception
from manila import manager
from manila import quota
from manila.share import rpcapi as share_rpc
from manila import utils

QUOTAS = quota.QUOTAS

from manila.i18n import _

LOG = log.getLogger(__name__)

backup_opts = [
    cfg.StrOpt(
        'backup_driver',
        default='manila.data.drivers.nfs.NFSBackupDriver',
        help='Driver to use for backups.'),
    cfg.StrOpt(
        'backup_share_mount_template',
        default='mount -vt %(proto)s %(options)s %(export)s %(path)s',
        help="The template for mounting shares during backup. Must specify "
             "the executable with all necessary parameters for the protocol "
             "supported. 'proto' template element may not be required if "
             "included in the command. 'export' and 'path' template elements "
             "are required. It is advisable to separate different commands "
             "per backend."),
    cfg.StrOpt(
        'backup_share_unmount_template',
        default='umount -v %(path)s',
        help="The template for unmounting shares during backup. Must "
             "specify the executable with all necessary parameters for the "
             "protocol supported. 'path' template element is required. It is "
             "advisable to separate different commands per backend."),
    cfg.ListOpt(
        'backup_ignore_files',
        default=['lost+found'],
        help="List of files and folders to be ignored when backing up "
             "shares. Items should be names (not including any path)."),
    cfg.Opt(
        'backup_protocol_access_mapping',
        type=types.Dict(types.List(types.String(), bounds=True)),
        default={'ip': ['nfs']},
        help="Protocol access mapping for backup. Should be a "
             "dictionary comprised of "
             "{'access_type1': ['share_proto1', 'share_proto2'],"
             " 'access_type2': ['share_proto2', 'share_proto3']}."),
]

data_opts = [
    cfg.StrOpt(
        'mount_tmp_location',
        default='/tmp/',
        help="Temporary path to create and mount shares during migration."),
    cfg.StrOpt(
        'backup_mount_tmp_location',
        default='/tmp/',
        help="Temporary path to create and mount backup during share backup."),
    cfg.BoolOpt(
        'check_hash',
        default=False,
        help="Chooses whether hash of each file should be checked on data "
             "copying."),
    cfg.IntOpt(
        'backup_continue_update_interval',
        default=10,
        help='This value, specified in seconds, determines how often '
             'the data manager will poll to perform the next steps of '
             'backup such as fetch the progress of backup.'),
    cfg.IntOpt(
        'restore_continue_update_interval',
        default=10,
        help='This value, specified in seconds, determines how often '
             'the data manager will poll to perform the next steps of '
             'restore such as fetch the progress of restore.')
]


CONF = cfg.CONF
CONF.register_opts(data_opts)
CONF.register_opts(backup_opts)


class DataManager(manager.Manager):
    """Receives requests to handle data and sends responses."""

    RPC_API_VERSION = '1.1'

    def __init__(self, service_name=None, *args, **kwargs):
        super(DataManager, self).__init__(*args, **kwargs)
        self.backup_driver = importutils.import_object(CONF.backup_driver)
        self.busy_tasks_shares = {}
        self.service_id = None

    def init_host(self, service_id=None):
        self.service_id = service_id
        ctxt = context.get_admin_context()
        shares = self.db.share_get_all(ctxt)
        for share in shares:
            if share['task_state'] in constants.BUSY_COPYING_STATES:
                self.db.share_update(
                    ctxt, share['id'],
                    {'task_state': constants.TASK_STATE_DATA_COPYING_ERROR})

    def migration_start(self, context, ignore_list, share_id,
                        share_instance_id, dest_share_instance_id,
                        connection_info_src, connection_info_dest):

        LOG.debug(
            "Received request to migrate share content from share instance "
            "%(instance_id)s to instance %(dest_instance_id)s.",
            {'instance_id': share_instance_id,
             'dest_instance_id': dest_share_instance_id})

        share_ref = self.db.share_get(context, share_id)
        share_instance_ref = self.db.share_instance_get(
            context, share_instance_id, with_share_data=True)

        share_rpcapi = share_rpc.ShareAPI()

        mount_path = CONF.mount_tmp_location

        try:
            copy = data_utils.Copy(
                os.path.join(mount_path, share_instance_id),
                os.path.join(mount_path, dest_share_instance_id),
                ignore_list, CONF.check_hash)

            info_src = {
                'share_id': share_ref['id'],
                'share_instance_id': share_instance_id,
                'mount': connection_info_src['mount'],
                'unmount': connection_info_src['unmount'],
                'access_mapping': connection_info_src.get(
                    'access_mapping', {}),
                'mount_point': os.path.join(mount_path,
                                            share_instance_id),
            }

            info_dest = {
                'share_id': None,
                'share_instance_id': dest_share_instance_id,
                'mount': connection_info_dest['mount'],
                'unmount': connection_info_dest['unmount'],
                'access_mapping': connection_info_dest.get(
                    'access_mapping', {}),
                'mount_point': os.path.join(mount_path,
                                            dest_share_instance_id),
            }

            self._copy_share_data(context, copy, info_src, info_dest)
        except exception.ShareDataCopyCancelled:
            share_rpcapi.migration_complete(
                context, share_instance_ref, dest_share_instance_id)
            return
        except Exception:
            self.db.share_update(
                context, share_id,
                {'task_state': constants.TASK_STATE_DATA_COPYING_ERROR})
            msg = _("Failed to copy contents from instance %(src)s to "
                    "instance %(dest)s.") % {'src': share_instance_id,
                                             'dest': dest_share_instance_id}
            LOG.exception(msg)
            share_rpcapi.migration_complete(
                context, share_instance_ref, dest_share_instance_id)
            raise exception.ShareDataCopyFailed(reason=msg)
        finally:
            self.busy_tasks_shares.pop(share_id, None)

        LOG.info(
            "Completed copy operation of migrating share content from share "
            "instance %(instance_id)s to instance %(dest_instance_id)s.",
            {'instance_id': share_instance_id,
             'dest_instance_id': dest_share_instance_id})

    def data_copy_cancel(self, context, share_id):
        LOG.debug("Received request to cancel data copy "
                  "of share %s.", share_id)
        copy = self.busy_tasks_shares.get(share_id)
        if copy:
            copy.cancel()
        else:
            msg = _("Data copy for migration of share %s cannot be cancelled"
                    " at this moment.") % share_id
            LOG.error(msg)
            raise exception.InvalidShare(reason=msg)

    def data_copy_get_progress(self, context, share_id):
        LOG.debug("Received request to get data copy information "
                  "of share %s.", share_id)
        copy = self.busy_tasks_shares.get(share_id)
        if copy:
            result = copy.get_progress()
            LOG.info("Obtained following data copy information "
                     "of share %(share)s: %(info)s.",
                     {'share': share_id,
                      'info': result})
            return result
        else:
            msg = _("Migration of share %s data copy progress cannot be "
                    "obtained at this moment.") % share_id
            LOG.error(msg)
            raise exception.InvalidShare(reason=msg)

    def _copy_share_data(self, context, copy, info_src, info_dest):
        """Copy share data between source and destination.

        e.g. During migration source and destination both are shares
        and during backup create, destination is backup location
        while during backup restore, source is backup location.
        1. Mount source and destination. Create access rules.
        2. Perform copy
        3. Unmount source and destination. Cleanup access rules.
        """
        mount_path = CONF.mount_tmp_location

        if info_src.get('share_id'):
            share_id = info_src['share_id']
        elif info_dest.get('share_id'):
            share_id = info_dest['share_id']
        else:
            msg = _("Share data copy failed because of undefined share.")
            LOG.exception(msg)
            raise exception.ShareDataCopyFailed(reason=msg)

        share_instance_src = None
        share_instance_dest = None
        if info_src['share_instance_id']:
            share_instance_src = self.db.share_instance_get(
                context, info_src['share_instance_id'], with_share_data=True)
        if info_dest['share_instance_id']:
            share_instance_dest = self.db.share_instance_get(
                context, info_dest['share_instance_id'], with_share_data=True)

        share = self.db.share_get(context, share_id)
        self.db.share_update(
            context, share['id'],
            {'task_state': constants.TASK_STATE_DATA_COPYING_STARTING})

        helper_src = helper.DataServiceHelper(context, self.db, share)
        helper_dest = helper_src

        if share_instance_src:
            access_ref_src = helper_src.allow_access_to_data_service(
                share_instance_src, info_src, share_instance_dest, info_dest)
            access_ref_dest = access_ref_src
        elif share_instance_dest:
            access_ref_src = helper_src.allow_access_to_data_service(
                share_instance_dest, info_dest, share_instance_src, info_src)
            access_ref_dest = access_ref_src

        def _call_cleanups(items):
            for item in items:
                if 'unmount_src' == item:
                    helper_src.cleanup_unmount_temp_folder(
                        info_src, mount_path)
                elif 'temp_folder_src' == item:
                    helper_src.cleanup_temp_folder(
                        mount_path, info_src['share_instance_id'])
                elif 'temp_folder_dest' == item:
                    helper_dest.cleanup_temp_folder(
                        mount_path, info_dest['share_instance_id'])
                elif 'access_src' == item and share_instance_src:
                    helper_src.cleanup_data_access(
                        access_ref_src, share_instance_src)
                elif 'access_dest' == item and share_instance_dest:
                    helper_dest.cleanup_data_access(
                        access_ref_dest, share_instance_dest)
        try:
            helper_src.mount_share_instance_or_backup(info_src, mount_path)
        except Exception:
            msg = _("Share data copy failed attempting to mount source "
                    "at %s.") % info_src['mount_point']
            LOG.exception(msg)
            _call_cleanups(['temp_folder_src', 'access_dest', 'access_src'])
            raise exception.ShareDataCopyFailed(reason=msg)

        try:
            helper_dest.mount_share_instance_or_backup(info_dest, mount_path)
        except Exception:
            msg = _("Share data copy failed attempting to mount destination "
                    "at %s.") % info_dest['mount_point']
            LOG.exception(msg)
            _call_cleanups(['temp_folder_dest', 'unmount_src',
                            'temp_folder_src', 'access_dest', 'access_src'])
            raise exception.ShareDataCopyFailed(reason=msg)

        self.busy_tasks_shares[share['id']] = copy
        self.db.share_update(
            context, share['id'],
            {'task_state': constants.TASK_STATE_DATA_COPYING_IN_PROGRESS})

        copied = False
        try:
            copy.run()
            self.db.share_update(
                context, share['id'],
                {'task_state': constants.TASK_STATE_DATA_COPYING_COMPLETING})
            if copy.get_progress()['total_progress'] == 100:
                copied = True
        except Exception:
            LOG.exception("Failed to copy data from source to destination "
                          "%(src)s to %(dest)s.",
                          {'src': info_src['mount_point'],
                           'dest': info_dest['mount_point']})

        try:
            helper_src.unmount_share_instance_or_backup(info_src,
                                                        mount_path)
        except Exception:
            LOG.exception("Could not unmount src %s after its data copy.",
                          info_src['mount_point'])

        try:
            helper_dest.unmount_share_instance_or_backup(info_dest,
                                                         mount_path)
        except Exception:
            LOG.exception("Could not unmount dest %s after its data copy.",
                          info_dest['mount_point'])

        try:
            if info_src['share_instance_id']:
                helper_src.deny_access_to_data_service(access_ref_src,
                                                       share_instance_src)
        except Exception:
            LOG.exception("Could not deny access to src instance %s after "
                          "its data copy.", info_src['share_instance_id'])

        try:
            if info_dest['share_instance_id']:
                helper_dest.deny_access_to_data_service(access_ref_dest,
                                                        share_instance_dest)
        except Exception:
            LOG.exception("Could not deny access to dest instance %s after "
                          "its data copy.", info_dest['share_instance_id'])

        if copy and copy.cancelled:
            self.db.share_update(
                context, share['id'],
                {'task_state': constants.TASK_STATE_DATA_COPYING_CANCELLED})
            LOG.warning("Copy of data from source "
                        "%(src)s to destination %(dest)s was cancelled.",
                        {'src': info_src['mount_point'],
                         'dest': info_dest['mount_point']})
            raise exception.ShareDataCopyCancelled()
        elif not copied:
            msg = _("Copying data from source %(src)s "
                    "to destination %(dest)s did not succeed.") % (
                        {'src': info_src['mount_point'],
                         'dest': info_dest['mount_point']})
            raise exception.ShareDataCopyFailed(reason=msg)

        self.db.share_update(
            context, share['id'],
            {'task_state': constants.TASK_STATE_DATA_COPYING_COMPLETED})

        LOG.debug("Copy of data from source %(src)s to destination "
                  "%(dest)s was successful.", {
                      'src': info_src['mount_point'],
                      'dest': info_dest['mount_point']})

    def create_backup(self, context, backup):
        share_id = backup['share_id']
        backup_id = backup['id']
        share = self.db.share_get(context, share_id)
        backup = self.db.share_backup_get(context, backup_id)

        self.db.share_backup_update(context, backup_id, {'host': self.host})

        LOG.info('Create backup started, backup: %(backup_id)s '
                 'share: %(share_id)s.',
                 {'backup_id': backup_id, 'share_id': share_id})

        try:
            self._run_backup(context, backup, share)
        except Exception as err:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create share backup %s by data driver.",
                          backup['id'])
                self.db.share_update(
                    context, share_id,
                    {'status': constants.STATUS_AVAILABLE})
                self.db.share_backup_update(
                    context, backup_id,
                    {'status': constants.STATUS_ERROR, 'fail_reason': err})
        self.db.share_update(
            context, share_id, {'status': constants.STATUS_AVAILABLE})
        self.db.share_backup_update(
            context, backup_id,
            {'status': constants.STATUS_AVAILABLE, 'progress': '100'})
        LOG.info("Created share backup %s successfully.", backup_id)

    @periodic_task.periodic_task(
        spacing=CONF.backup_continue_update_interval)
    def create_backup_continue(self, context):
        filters = {
            'status': constants.STATUS_CREATING,
            'host': self.host,
            'topic': CONF.data_topic
        }
        backups = self.db.share_backups_get_all(context, filters)

        for backup in backups:
            backup_id = backup['id']
            share_id = backup['share_id']
            result = {}
            try:
                result = self.data_copy_get_progress(context, share_id)
                progress = result.get('total_progress', '0')
                backup_values = {'progress': progress}
                if progress == '100':
                    self.db.share_update(
                        context, share_id,
                        {'status': constants.STATUS_AVAILABLE})
                    backup_values.update(
                        {'status': constants.STATUS_AVAILABLE})
                    LOG.info("Created share backup %s successfully.",
                             backup_id)
                self.db.share_backup_update(context, backup_id, backup_values)
            except Exception:
                LOG.warning("Failed to get progress of share %(share)s "
                            "backing up in share_backup %(backup).",
                            {'share': share_id, 'backup': backup_id})
                self.db.share_update(
                    context, share_id,
                    {'status': constants.STATUS_AVAILABLE})
                self.db.share_backup_update(
                    context, backup_id,
                    {'status': constants.STATUS_ERROR, 'progress': '0'})

    def _get_share_mount_info(self, share_instance):
        mount_template = CONF.backup_share_mount_template

        path = next((x['path'] for x in share_instance['export_locations']
                    if x['is_admin_only']), None)
        if not path:
            path = share_instance['export_locations'][0]['path']

        format_args = {
            'proto': share_instance['share_proto'].lower(),
            'export': path,
            'path': '%(path)s',
            'options': '%(options)s',
        }

        unmount_template = CONF.backup_share_unmount_template
        mount_info = {
            'mount': mount_template % format_args,
            'unmount': unmount_template,
        }
        return mount_info

    def _get_backup_access_mapping(self, share):
        mapping = CONF.backup_protocol_access_mapping
        result = {}
        share_proto = share['share_proto'].lower()
        for access_type, protocols in mapping.items():
            if share_proto in [y.lower() for y in protocols]:
                result[access_type] = result.get(access_type, [])
                result[access_type].append(share_proto)
        return result

    def _run_backup(self, context, backup, share):
        share_instance_id = share.instance.get('id')
        share_instance = self.db.share_instance_get(
            context, share_instance_id, with_share_data=True)

        access_mapping = self._get_backup_access_mapping(share)
        ignore_list = CONF.backup_ignore_files
        mount_path = CONF.mount_tmp_location
        backup_mount_path = CONF.backup_mount_tmp_location

        mount_info = self._get_share_mount_info(share_instance)
        dest_backup_info = self.backup_driver.get_backup_info(backup)

        dest_backup_mount_point = os.path.join(backup_mount_path, backup['id'])
        backup_folder = os.path.join(dest_backup_mount_point, backup['id'])

        try:
            copy = data_utils.Copy(
                os.path.join(mount_path, share_instance_id),
                backup_folder,
                ignore_list)

            info_src = {
                'share_id': share['id'],
                'share_instance_id': share_instance_id,
                'mount': mount_info['mount'],
                'unmount': mount_info['unmount'],
                'mount_point': os.path.join(mount_path, share_instance_id),
                'access_mapping': access_mapping
            }

            info_dest = {
                'share_id': None,
                'share_instance_id': None,
                'backup': True,
                'backup_id': backup['id'],
                'mount': dest_backup_info['mount'],
                'unmount': dest_backup_info['unmount'],
                'mount_point': dest_backup_mount_point,
                'access_mapping': access_mapping
            }
            self._copy_share_data(context, copy, info_src, info_dest)
            self.db.share_update(context, share['id'], {'task_state': None})
        except Exception:
            self.db.share_update(
                context, share['id'],
                {'task_state': constants.TASK_STATE_DATA_COPYING_ERROR})
            msg = _("Failed to copy contents from share %(src)s to "
                    "backup %(dest)s.") % (
                        {'src': share_instance_id, 'dest': backup['id']})
            LOG.exception(msg)
            raise exception.ShareDataCopyFailed(reason=msg)
        finally:
            self.busy_tasks_shares.pop(share['id'], None)

    def delete_backup(self, context, backup):
        backup_id = backup['id']
        LOG.info('Delete backup started, backup: %s.', backup_id)

        backup = self.db.share_backup_get(context, backup_id)
        try:
            dest_backup_info = self.backup_driver.get_backup_info(backup)
            backup_mount_path = CONF.backup_mount_tmp_location
            mount_point = os.path.join(backup_mount_path, backup['id'])
            backup_folder = os.path.join(mount_point, backup['id'])
            if not os.path.exists(backup_folder):
                os.makedirs(backup_folder)
            if not os.path.exists(backup_folder):
                raise exception.NotFound("Path %s could not be "
                                         "found." % backup_folder)

            mount_template = dest_backup_info['mount']
            unmount_template = dest_backup_info['unmount']
            mount_command = mount_template % {'path': mount_point}
            unmount_command = unmount_template % {'path': mount_point}
            utils.execute(*(mount_command.split()), run_as_root=True)

            # backup_folder should exist after mount, else backup is
            # already deleted
            if os.path.exists(backup_folder):
                for filename in os.listdir(backup_folder):
                    if filename in CONF.backup_ignore_files:
                        continue
                    file_path = os.path.join(backup_folder, filename)
                    try:
                        if (os.path.isfile(file_path) or
                                os.path.islink(file_path)):
                            os.unlink(file_path)
                        elif os.path.isdir(file_path):
                            shutil.rmtree(file_path)
                    except Exception as e:
                        LOG.debug("Failed to delete %(file_path)s. Reason: "
                                  "%(err)s", {'file_path': file_path,
                                              'err': e})
                shutil.rmtree(backup_folder)
            utils.execute(*(unmount_command.split()), run_as_root=True)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to delete share backup %s.", backup['id'])
                self.db.share_backup_update(
                    context, backup['id'],
                    {'status': constants.STATUS_ERROR_DELETING})

        try:
            reserve_opts = {
                'backups': -1,
                'backup_gigabytes': -backup['size'],
            }
            reservations = QUOTAS.reserve(
                context, project_id=backup['project_id'], **reserve_opts)
        except Exception as e:
            reservations = None
            LOG.warning("Failed to update backup quota for %(pid)s: %(err)s.",
                        {'pid': backup['project_id'], 'err': e})
            raise

        if reservations:
            QUOTAS.commit(context, reservations,
                          project_id=backup['project_id'])

        self.db.share_backup_delete(context, backup_id)
        LOG.info("Share backup %s deleted successfully.", backup_id)

    def restore_backup(self, context, backup, share_id):
        backup_id = backup['id']
        LOG.info('Restore backup started, backup: %(backup_id)s '
                 'share: %(share_id)s.',
                 {'backup_id': backup['id'], 'share_id': share_id})

        share = self.db.share_get(context, share_id)
        backup = self.db.share_backup_get(context, backup_id)

        try:
            self._run_restore(context, backup, share)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to restore backup %(backup)s to share "
                          "%(share)s by data driver.",
                          {'backup': backup['id'], 'share': share_id})
                self.db.share_update(
                    context, share_id,
                    {'status': constants.STATUS_BACKUP_RESTORING_ERROR})
                self.db.share_backup_update(
                    context, backup_id,
                    {'status': constants.STATUS_AVAILABLE})
        self.db.share_update(
            context, share_id, {'status': constants.STATUS_AVAILABLE})
        self.db.share_backup_update(
            context, backup_id,
            {'status': constants.STATUS_AVAILABLE, 'restore_progress': '100'})
        LOG.info("Share backup %s restored successfully.", backup_id)

    @periodic_task.periodic_task(
        spacing=CONF.restore_continue_update_interval)
    def restore_backup_continue(self, context):
        filters = {
            'status': constants.STATUS_RESTORING,
            'host': self.host,
            'topic': CONF.data_topic
        }
        backups = self.db.share_backups_get_all(context, filters)
        for backup in backups:
            backup_id = backup['id']
            try:
                filters = {
                    'source_backup_id': backup_id,
                }
                shares = self.db.share_get_all(context, filters)
            except Exception:
                LOG.warning('Failed to get shares for backup %s', backup_id)
                continue

            for share in shares:
                if share['status'] != constants.STATUS_BACKUP_RESTORING:
                    continue

                share_id = share['id']
                result = {}
                try:
                    result = self.data_copy_get_progress(context, share_id)
                    progress = result.get('total_progress', '0')
                    backup_values = {'restore_progress': progress}
                    if progress == '100':
                        self.db.share_update(
                            context, share_id,
                            {'status': constants.STATUS_AVAILABLE})
                        backup_values.update(
                            {'status': constants.STATUS_AVAILABLE})
                        LOG.info("Share backup %s restored successfully.",
                                 backup_id)
                    self.db.share_backup_update(context, backup_id,
                                                backup_values)
                except Exception:
                    LOG.exception("Failed to get progress of share_backup "
                                  "%(backup)s restoring in share %(share).",
                                  {'share': share_id, 'backup': backup_id})
                    self.db.share_update(
                        context, share_id,
                        {'status': constants.STATUS_BACKUP_RESTORING_ERROR})
                    self.db.share_backup_update(
                        context, backup_id,
                        {'status': constants.STATUS_AVAILABLE,
                         'restore_progress': '0'})

    def _run_restore(self, context, backup, share):
        share_instance_id = share.instance.get('id')
        share_instance = self.db.share_instance_get(
            context, share_instance_id, with_share_data=True)

        access_mapping = self._get_backup_access_mapping(share)
        mount_path = CONF.mount_tmp_location
        backup_mount_path = CONF.backup_mount_tmp_location
        ignore_list = CONF.backup_ignore_files

        mount_info = self._get_share_mount_info(share_instance)
        src_backup_info = self.backup_driver.get_backup_info(backup)

        src_backup_mount_point = os.path.join(backup_mount_path, backup['id'])
        backup_folder = os.path.join(src_backup_mount_point, backup['id'])

        try:
            copy = data_utils.Copy(
                backup_folder,
                os.path.join(mount_path, share_instance_id),
                ignore_list)

            info_src = {
                'share_id': None,
                'share_instance_id': None,
                'restore': True,
                'backup_id': backup['id'],
                'mount': src_backup_info['mount'],
                'unmount': src_backup_info['unmount'],
                'mount_point': src_backup_mount_point,
                'access_mapping': access_mapping
            }

            info_dest = {
                'share_id': share['id'],
                'share_instance_id': share_instance_id,
                'mount': mount_info['mount'],
                'unmount': mount_info['unmount'],
                'mount_point': os.path.join(mount_path, share_instance_id),
                'access_mapping': access_mapping
            }

            self._copy_share_data(context, copy, info_src, info_dest)
            self.db.share_update(context, share['id'], {'task_state': None})
        except Exception:
            self.db.share_update(
                context, share['id'],
                {'task_state': constants.TASK_STATE_DATA_COPYING_ERROR})
            msg = _("Failed to copy/restore contents from backup %(src)s "
                    "to share %(dest)s.") % (
                        {'src': backup['id'], 'dest': share_instance_id})
            LOG.exception(msg)
            raise exception.ShareDataCopyFailed(reason=msg)
        finally:
            self.busy_tasks_shares.pop(share['id'], None)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ddt
from oslo_config import cfg
from unittest import mock
from webob import exc

from manila.api.openstack import api_version_request as api_version
from manila.api.v2 import share_backups
from manila.common import constants
from manila import context
from manila import exception
from manila import policy
from manila import share
from manila import test
from manila.tests.api import fakes
from manila.tests import db_utils
from manila.tests import fake_share

CONF = cfg.CONF


@ddt.ddt
class ShareBackupsApiTest(test.TestCase):
    """Share backups API Test Cases."""
    def setUp(self):
        super(ShareBackupsApiTest, self).setUp()
        self.controller = share_backups.ShareBackupController()
        self.resource_name = self.controller.resource_name
        self.api_version = share_backups.MIN_SUPPORTED_API_VERSION
        self.backups_req = fakes.HTTPRequest.blank(
            '/share-backups', version=self.api_version,
            experimental=True)
        self.member_context = context.RequestContext('fake', 'fake')
        self.backups_req.environ['manila.context'] = self.member_context
        self.backups_req_admin = fakes.HTTPRequest.blank(
            '/share-backups', version=self.api_version,
            experimental=True, use_admin_context=True)
        self.admin_context = self.backups_req_admin.environ['manila.context']
        self.mock_policy_check = self.mock_object(policy, 'check_policy')

    def _get_context(self, role):
        return getattr(self, '%s_context' % role)

    def _create_backup_get_req(self, **kwargs):
        if 'status' not in kwargs:
            kwargs['status'] = constants.STATUS_AVAILABLE
        backup = db_utils.create_share_backup(**kwargs)
        req = fakes.HTTPRequest.blank(
            '/v2/fake/share-backups/%s/action' % backup['id'],
            version=self.api_version)
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        req.headers['X-Openstack-Manila-Api-Version'] = self.api_version
        req.headers['X-Openstack-Manila-Api-Experimental'] = True

        return backup, req

    def _get_fake_backup(self, admin=False, summary=False,
                         apiversion=share_backups.MIN_SUPPORTED_API_VERSION,
                         **values):
        backup = fake_share.fake_backup(**values)
        backup['updated_at'] = '2016-06-12T19:57:56.506805'
        expected_keys = {'id', 'share_id', 'status'}
        expected_backup = {key: backup[key] for key in backup if key
                           in expected_keys}
        expected_backup.update({'name': backup.get('display_name')})

        if not summary:
            expected_backup.update({
                'id': backup.get('id'),
                'share_id': backup.get('share_id'),
                'status': backup.get('status'),
                'description': backup.get('display_description'),
                'size': backup.get('size'),
                'created_at': backup.get('created_at'),
                'updated_at': backup.get('updated_at'),
                'availability_zone': backup.get('availability_zone'),
                'progress': backup.get('progress'),
                'restore_progress': backup.get('restore_progress'),
            })
            if self.is_microversion_ge(apiversion, '2.85'):
                expected_backup.update({
                    'backup_type': backup.get('backup_type'),
                })

            if admin:
                expected_backup.update({
                    'host': backup.get('host'),
                    'topic': backup.get('topic'),
                })

        return backup, expected_backup

    def test_list_backups_summary(self):
        fake_backup, expected_backup = self._get_fake_backup(summary=True)
        self.mock_object(share_backups.db, 'share_backups_get_all',
                         mock.Mock(return_value=[fake_backup]))

        res_dict = self.controller.index(self.backups_req)
        self.assertEqual([expected_backup], res_dict['share_backups'])
        self.mock_policy_check.assert_called_once_with(
            self.member_context, self.resource_name, 'get_all')

    def test_list_backups_summary_with_share_id(self):
        fake_backup, expected_backup = self._get_fake_backup(summary=True)
        self.mock_object(share.API, 'get',
                         mock.Mock(return_value={'id': 'FAKE_SHAREID'}))
        self.mock_object(share_backups.db, 'share_backups_get_all',
                         mock.Mock(return_value=[fake_backup]))
        req = fakes.HTTPRequest.blank(
            '/share-backups?share_id=FAKE_SHARE_ID',
            version=self.api_version, experimental=True)
        req_context = req.environ['manila.context']

        res_dict = self.controller.index(req)

        self.assertEqual([expected_backup], res_dict['share_backups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    @ddt.data(True, False)
    def test_list_backups_detail(self, is_admin):
        fake_backup, expected_backup = self._get_fake_backup(admin=is_admin)
        self.mock_object(share_backups.db, 'share_backups_get_all',
                         mock.Mock(return_value=[fake_backup]))

        req = self.backups_req if not is_admin else self.backups_req_admin
        req_context = req.environ['manila.context']

        res_dict = self.controller.detail(req)

        self.assertEqual([expected_backup], res_dict['share_backups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_list_share_backups_detail_with_limit(self):
        fake_backup_1, expected_backup_1 = self._get_fake_backup()
        fake_backup_2, expected_backup_2 = self._get_fake_backup(
            id="fake_id2")
        self.mock_object(
            share_backups.db, 'share_backups_get_all',
            mock.Mock(return_value=[fake_backup_1]))
        req = fakes.HTTPRequest.blank('/share-backups?limit=1',
                                      version=self.api_version,
                                      experimental=True)
        req_context = req.environ['manila.context']

        res_dict = self.controller.detail(req)

        self.assertEqual(1, len(res_dict['share_backups']))
        self.assertEqual([expected_backup_1], res_dict['share_backups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_list_share_backups_detail_with_limit_and_offset(self):
        fake_backup_1, expected_backup_1 = self._get_fake_backup()
        fake_backup_2, expected_backup_2 = self._get_fake_backup(
            id="fake_id2")
        self.mock_object(
            share_backups.db, 'share_backups_get_all',
            mock.Mock(return_value=[fake_backup_2]))
        req = fakes.HTTPRequest.blank(
            '/share-backups/detail?limit=1&offset=1',
            version=self.api_version, experimental=True)
        req_context = req.environ['manila.context']

        res_dict = self.controller.detail(req)

        self.assertEqual(1, len(res_dict['share_backups']))
        self.assertEqual([expected_backup_2], res_dict['share_backups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_list_share_backups_detail_invalid_share(self):
        self.mock_object(share_backups.db, 'share_backups_get_all',
                         mock.Mock(side_effect=exception.NotFound))
        mock__view_builder_call = self.mock_object(
            share_backups.backup_view.BackupViewBuilder,
            'detail_list')
        req = self.backups_req
        req.GET['share_id'] = 'FAKE_SHARE_ID'

        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.detail, req)
        self.assertFalse(mock__view_builder_call.called)
        self.mock_policy_check.assert_called_once_with(
            self.member_context, self.resource_name, 'get_all')

    @ddt.data(share_backups.MIN_SUPPORTED_API_VERSION,
              api_version._MAX_API_VERSION)
    def test_list_share_backups_detail(self, apiversion):
        fake_backup, expected_backup = self._get_fake_backup(
            apiversion=apiversion,
        )
        self.mock_object(share.API, 'get',
                         mock.Mock(return_value={'id': 'FAKE_SHAREID'}))
        self.mock_object(share_backups.db, 'share_backups_get_all',
                         mock.Mock(return_value=[fake_backup]))
        req = fakes.HTTPRequest.blank(
            '/share-backups?share_id=FAKE_SHARE_ID',
            version=apiversion, experimental=True)
        req.environ['manila.context'] = (
            self.member_context)
        req_context = req.environ['manila.context']

        res_dict = self.controller.detail(req)

        self.assertEqual([expected_backup], res_dict['share_backups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_list_share_backups_with_limit(self):
        fake_backup_1, expected_backup_1 = self._get_fake_backup()
        fake_backup_2, expected_backup_2 = self._get_fake_backup(
            id="fake_id2")

        self.mock_object(share.API, 'get',
                         mock.Mock(return_value={'id': 'FAKE_SHAREID'}))
        self.mock_object(
            share_backups.db, 'share_backups_get_all',
            mock.Mock(return_value=[fake_backup_1]))
        req = fakes.HTTPRequest.blank(
            '/share-backups?share_id=FAKE_SHARE_ID&limit=1',
            version=self.api_version, experimental=True)
        req_context = req.environ['manila.context']

        res_dict = self.controller.detail(req)

        self.assertEqual(1, len(res_dict['share_backups']))
        self.assertEqual([expected_backup_1], res_dict['share_backups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_list_share_backups_with_limit_and_offset(self):
        fake_backup_1, expected_backup_1 = self._get_fake_backup()
        fake_backup_2, expected_backup_2 = self._get_fake_backup(
            id="fake_id2")
        self.mock_object(share.API, 'get',
                         mock.Mock(return_value={'id': 'FAKE_SHAREID'}))
        self.mock_object(
            share_backups.db, 'share_backups_get_all',
            mock.Mock(return_value=[fake_backup_2]))
        req = fakes.HTTPRequest.blank(
            '/share-backups?share_id=FAKE_SHARE_ID&limit=1&offset=1',
            version=self.api_version, experimental=True)
        req_context = req.environ['manila.context']

        res_dict = self.controller.detail(req)

        self.assertEqual(1, len(res_dict['share_backups']))
        self.assertEqual([expected_backup_2], res_dict['share_backups'])
        self.mock_policy_check.assert_called_once_with(
            req_context, self.resource_name, 'get_all')

    def test_show(self):
        fake_backup, expected_backup = self._get_fake_backup()
        self.mock_object(
            share_backups.db, 'share_backup_get',
            mock.Mock(return_value=fake_backup))

        req = self.backups_req
        res_dict = self.controller.show(req, fake_backup.get('id'))

        self.assertEqual(expected_backup, res_dict['share_backup'])

    def test_show_no_backup(self):
        mock__view_builder_call = self.mock_object(
            share_backups.backup_view.BackupViewBuilder, 'detail')
        fake_exception = exception.ShareBackupNotFound(
            backup_id='FAKE_backup_ID')
        self.mock_object(share_backups.db, 'share_backup_get', mock.Mock(
            side_effect=fake_exception))

        self.assertRaises(exc.HTTPNotFound,
                          self.controller.show,
                          self.backups_req,
                          'FAKE_backup_ID')
        self.assertFalse(mock__view_builder_call.called)

    def test_create_invalid_body(self):
        body = {}
        mock__view_builder_call = self.mock_object(
            share_backups.backup_view.BackupViewBuilder,
            'detail_list')

        self.assertRaises(exc.HTTPUnprocessableEntity,
                          self.controller.create,
                          self.backups_req, body)
        self.assertEqual(0, mock__view_builder_call.call_count)

    def test_create_no_share_id(self):
        body = {
            'share_backup': {
                'share_id': None,
                'availability_zone': None,
            }
        }
        mock__view_builder_call = self.mock_object(
            share_backups.backup_view.BackupViewBuilder,
            'detail_list')
        self.mock_object(share_backups.db, 'share_get',
                         mock.Mock(side_effect=exception.NotFound))

        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.create,
                          self.backups_req, body)
        self.assertFalse(mock__view_builder_call.called)

    def test_create_invalid_share_id(self):
        body = {
            'share_backup': {
                'share_id': None,
            }
        }
        mock__view_builder_call = self.mock_object(
            share_backups.backup_view.BackupViewBuilder,
            'detail_list')
        self.mock_object(share.API, 'get',
                         mock.Mock(side_effect=exception.NotFound))

        self.assertRaises(exc.HTTPBadRequest,
                          self.controller.create,
                          self.backups_req, body)
        self.assertFalse(mock__view_builder_call.called)

    @ddt.data(exception.InvalidBackup, exception.ShareBusyException)
    def test_create_exception_path(self, exception_type):
        fake_backup, _ = self._get_fake_backup()
        mock__view_builder_call = self.mock_object(
            share_backups.backup_view.BackupViewBuilder,
            'detail_list')
        body = {
            'share_backup': {
                'share_id': 'FAKE_SHAREID',
            }
        }
        exc_args = {'id': 'xyz', 'reason': 'abc'}
        self.mock_object(share.API, 'get',
                         mock.Mock(return_value={'id': 'FAKE_SHAREID'}))
        self.mock_object(share.API, 'create_share_backup',
                         mock.Mock(side_effect=exception_type(**exc_args)))

        if exception_type == exception.InvalidBackup:
            expected_exception = exc.HTTPBadRequest
        else:
            expected_exception = exc.HTTPConflict
        self.assertRaises(expected_exception,
                          self.controller.create,
                          self.backups_req, body)
        self.assertFalse(mock__view_builder_call.called)

    def test_create(self):
        fake_backup, expected_backup = self._get_fake_backup()
        body = {
            'share_backup': {
                'share_id': 'FAKE_SHAREID',
            }
        }
        self.mock_object(share.API, 'get',
                         mock.Mock(return_value={'id': 'FAKE_SHAREID'}))
        self.mock_object(share.API, 'create_share_backup',
                         mock.Mock(return_value=fake_backup))

        req = self.backups_req
        res_dict = self.controller.create(req, body)
        self.assertEqual(expected_backup, res_dict['share_backup'])

    def test_delete_invalid_backup(self):
        fake_exception = exception.ShareBackupNotFound(
            backup_id='FAKE_backup_ID')
        self.mock_object(share_backups.db, 'share_backup_get',
                         mock.Mock(side_effect=fake_exception))
        mock_delete_backup_call = self.mock_object(
            share.API, 'delete_share_backup')

        self.assertRaises(
            exc.HTTPNotFound, self.controller.delete,
            self.backups_req, 'FAKE_backup_ID')
        self.assertFalse(mock_delete_backup_call.called)

    def test_delete_exception(self):
        fake_backup_1 = self._get_fake_backup(
            share_id='FAKE_SHARE_ID',
            status=constants.STATUS_BACKUP_CREATING)[0]
        fake_backup_2 = self._get_fake_backup(
            share_id='FAKE_SHARE_ID',
            status=constants.STATUS_BACKUP_CREATING)[0]
        exception_type = exception.InvalidBackup(reason='xyz')
        self.mock_object(share_backups.db, 'share_backup_get',
                         mock.Mock(return_value=fake_backup_1))
        self.mock_object(
            share_backups.db, 'share_backups_get_all',
            mock.Mock(return_value=[fake_backup_1, fake_backup_2]))
        self.mock_object(share.API, 'delete_share_backup',
                         mock.Mock(side_effect=exception_type))

        self.assertRaises(exc.HTTPBadRequest, self.controller.delete,
                          self.backups_req, 'FAKE_backup_ID')

    def test_delete(self):
        fake_backup = self._get_fake_backup(
            share_id='FAKE_SHARE_ID',
            status=constants.STATUS_AVAILABLE)[0]
        self.mock_object(share_backups.db, 'share_backup_get',
                         mock.Mock(return_value=fake_backup))
        self.mock_object(share.API, 'delete_share_backup')

        resp = self.controller.delete(
            self.backups_req, 'FAKE_backup_ID')

        self.assertEqual(202, resp.status_code)

    def test_restore_invalid_backup_id(self):
        body = {'restore': None}
        fake_exception = exception.ShareBackupNotFound(
            backup_id='FAKE_BACKUP_ID')
        self.mock_object(share.API, 'restore',
                         mock.Mock(side_effect=fake_exception))

        self.assertRaises(exc.HTTPNotFound,
                          self.controller.restore,
                          self.backups_req,
                          'FAKE_BACKUP_ID', body)

    def test_restore(self):
        body = {'restore': {'share_id': 'fake_id'}}
        fake_backup = self._get_fake_backup(
            share_id='FAKE_SHARE_ID',
            status=constants.STATUS_AVAILABLE)[0]
        self.mock_object(share_backups.db, 'share_backup_get',
                         mock.Mock(return_value=fake_backup))

        fake_backup_restore = {
            'share_id': 'FAKE_SHARE_ID',
            'backup_id': fake_backup['id'],
        }
        mock_api_restore_backup_call = self.mock_object(
            share.API, 'restore_share_backup',
            mock.Mock(return_value=fake_backup_restore))
        self.mock_object(share.API, 'get',
                         mock.Mock(return_value={'id': 'FAKE_SHAREID'}))

        resp = self.controller.restore(self.backups_req,
                                       fake_backup['id'], body)

        self.assertEqual(fake_backup_restore, resp['restore'])
        self.assertTrue(mock_api_restore_backup_call.called)

    def test_update(self):
        fake_backup = self._get_fake_backup(
            share_id='FAKE_SHARE_ID',
            status=constants.STATUS_AVAILABLE)[0]
        self.mock_object(share_backups.db, 'share_backup_get',
                         mock.Mock(return_value=fake_backup))

        body = {'share_backup': {'name': 'backup1'}}
        fake_backup_update = {
            'share_id': 'FAKE_SHARE_ID',
            'backup_id': fake_backup['id'],
            'display_name': 'backup1'
        }
        mock_api_update_backup_call = self.mock_object(
            share.API, 'update_share_backup',
            mock.Mock(return_value=fake_backup_update))

        resp = self.controller.update(self.backups_req,
                                      fake_backup['id'], body)

        self.assertEqual(fake_backup_update['display_name'],
                         resp['share_backup']['name'])
        self.assertTrue(mock_api_update_backup_call.called)

    @ddt.data('index', 'detail')
    def test_policy_not_authorized(self, method_name):

        method = getattr(self.controller, method_name)
        arguments = {
            'id': 'FAKE_backup_ID',
            'body': {'FAKE_KEY': 'FAKE_VAL'},
        }
        if method_name in ('index', 'detail'):
            arguments.clear()

        noauthexc = exception.PolicyNotAuthorized(action=method)

        with mock.patch.object(
                policy, 'check_policy', mock.Mock(side_effect=noauthexc)):

            self.assertRaises(
                exc.HTTPForbidden, method, self.backups_req, **arguments)

    @ddt.data('index', 'detail', 'show', 'create', 'delete')
    def test_upsupported_microversion(self, method_name):

        unsupported_microversions = ('1.0', '2.2', '2.18')
        method = getattr(self.controller, method_name)
        arguments = {
            'id': 'FAKE_BACKUP_ID',
            'body': {'FAKE_KEY': 'FAKE_VAL'},
        }
        if method_name in ('index', 'detail'):
            arguments.clear()

        for microversion in unsupported_microversions:
            req = fakes.HTTPRequest.blank(
                '/share-backups', version=microversion,
                experimental=True)
            self.assertRaises(exception.VersionNotFoundForAPIMethod,
                              method, req, **arguments)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from webob import exc

from manila.api import common
from manila.api.openstack import wsgi
from manila import db
from manila import exception
from manila.i18n import _
from manila import policy


class MetadataController(object):
    """An abstract metadata controller resource."""

    # From db, ensure it exists
    resource_get = {
        "share": "share_get",
        "share_snapshot": "share_snapshot_get",
        "share_network_subnet": "share_network_subnet_get",
    }

    resource_metadata_get = {
        "share": "share_metadata_get",
        "share_snapshot": "share_snapshot_metadata_get",
        "share_network_subnet": "share_network_subnet_metadata_get",
    }

    resource_metadata_get_item = {
        "share": "share_metadata_get_item",
        "share_snapshot": "share_snapshot_metadata_get_item",
        "share_network_subnet": "share_network_subnet_metadata_get_item",
    }

    resource_metadata_update = {
        "share": "share_metadata_update",
        "share_snapshot": "share_snapshot_metadata_update",
        "share_network_subnet": "share_network_subnet_metadata_update",
    }

    resource_metadata_update_item = {
        "share": "share_metadata_update_item",
        "share_snapshot": "share_snapshot_metadata_update_item",
        "share_network_subnet": "share_network_subnet_metadata_update_item",
    }

    resource_metadata_delete = {
        "share": "share_metadata_delete",
        "share_snapshot": "share_snapshot_metadata_delete",
        "share_network_subnet": "share_network_subnet_metadata_delete",
    }

    resource_policy_get = {
        'share': 'get',
        'share_snapshot': 'get_snapshot',
        'share_network_subnet': 'show',
    }

    def __init__(self):
        super(MetadataController, self).__init__()
        self.resource_name = None

    def _get_resource(self, context, resource_id,
                      for_modification=False, parent_id=None):
        if self.resource_name in ['share', 'share_network_subnet']:
            # we would allow retrieving some "public" resources
            # across project namespaces except share snapshots,
            # project_only=True is hard coded
            kwargs = {}
        else:
            kwargs = {'project_only': True}
        try:
            get_res_method = getattr(
                db, self.resource_get[self.resource_name])
            if parent_id is not None:
                kwargs["parent_id"] = parent_id
            res = get_res_method(context, resource_id, **kwargs)

            get_policy = self.resource_policy_get[self.resource_name]
            if res.get('is_public') is False:
                authorized = policy.check_policy(context,
                                                 self.resource_name,
                                                 get_policy,
                                                 res,
                                                 do_raise=False)
                if not authorized:
                    # Raising NotFound to prevent existence detection
                    raise exception.NotFound()
            elif for_modification:
                # a public resource's metadata can be viewed, but not
                # modified by non owners
                policy.check_policy(context,
                                    self.resource_name,
                                    get_policy,
                                    res)
        except exception.NotFound:
            msg = _('%s not found.' % self.resource_name.capitalize())
            raise exc.HTTPNotFound(explanation=msg)
        return res

    def _get_metadata(self, context, resource_id, parent_id=None):

        self._get_resource(context, resource_id, parent_id=parent_id)
        get_metadata_method = getattr(
            db, self.resource_metadata_get[self.resource_name])

        result = get_metadata_method(context, resource_id)

        return result

    @wsgi.response(200)
    def _index_metadata(self, req, resource_id, parent_id=None):
        context = req.environ['manila.context']
        metadata = self._get_metadata(context, resource_id,
                                      parent_id=parent_id)

        return {'metadata': metadata}

    @wsgi.response(200)
    def _create_metadata(self, req, resource_id, body, parent_id=None):
        """Returns the new metadata item created."""

        context = req.environ['manila.context']
        try:
            metadata = body['metadata']
            common.check_metadata_properties(metadata)
        except (KeyError, TypeError):
            msg = _("Malformed request body")
            raise exc.HTTPBadRequest(explanation=msg)
        except exception.InvalidMetadata as error:
            raise exc.HTTPBadRequest(explanation=error.msg)
        except exception.InvalidMetadataSize as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        self._get_resource(context, resource_id,
                           for_modification=True, parent_id=parent_id)

        create_metadata_method = getattr(
            db, self.resource_metadata_update[self.resource_name])
        result = create_metadata_method(context, resource_id, metadata,
                                        delete='False')

        return {'metadata': result}

    def _update_metadata_item(self, req, resource_id, body, key,
                              parent_id=None):
        """Updates the specified metadata item."""

        context = req.environ['manila.context']
        try:
            meta_item = body['metadata']
            common.check_metadata_properties(meta_item)
        except (TypeError, KeyError):
            expl = _('Malformed request body')
            raise exc.HTTPBadRequest(explanation=expl)
        except exception.InvalidMetadata as error:
            raise exc.HTTPBadRequest(explanation=error.msg)
        except exception.InvalidMetadataSize as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        if key not in meta_item:
            expl = _('Request body and URI mismatch')
            raise exc.HTTPBadRequest(explanation=expl)
        if len(meta_item) > 1:
            expl = _('Request body contains too many items')
            raise exc.HTTPBadRequest(explanation=expl)
        self._get_resource(context, resource_id,
                           for_modification=True, parent_id=parent_id)

        update_metadata_item_method = getattr(
            db, self.resource_metadata_update_item[self.resource_name])
        result = update_metadata_item_method(context, resource_id, meta_item)

        return {'metadata': result}

    @wsgi.response(200)
    def _update_all_metadata(self, req, resource_id, body, parent_id=None):
        """Deletes existing metadata, and returns the updated metadata."""

        context = req.environ['manila.context']
        try:
            metadata = body['metadata']
            common.check_metadata_properties(metadata)
        except (TypeError, KeyError):
            expl = _('Malformed request body')
            raise exc.HTTPBadRequest(explanation=expl)
        except exception.InvalidMetadata as error:
            raise exc.HTTPBadRequest(explanation=error.msg)
        except exception.InvalidMetadataSize as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        self._get_resource(context, resource_id,
                           for_modification=True, parent_id=parent_id)
        meta_ref = self._get_metadata(context, resource_id,
                                      parent_id=parent_id)

        for key in meta_ref:
            delete_metadata_method = getattr(
                db, self.resource_metadata_delete[self.resource_name])
            delete_metadata_method(context, resource_id, key)

        update_metadata_method = getattr(
            db, self.resource_metadata_update[self.resource_name])
        new_metadata = update_metadata_method(context, resource_id,
                                              metadata, delete='False')
        return {'metadata': new_metadata}

    @wsgi.response(200)
    def _show_metadata(self, req, resource_id, key, parent_id=None):
        """Return metadata item."""

        context = req.environ['manila.context']
        self._get_resource(context, resource_id,
                           for_modification=False, parent_id=parent_id)
        get_metadata_item_method = getattr(
            db, self.resource_metadata_get_item[self.resource_name])
        item = get_metadata_item_method(context, resource_id, key)

        return {'meta': {key: item[key]}}

    @wsgi.response(200)
    def _delete_metadata(self, req, resource_id, key, parent_id=None):
        """Deletes existing metadata item."""

        context = req.environ['manila.context']
        self._get_resource(context, resource_id,
                           for_modification=True, parent_id=parent_id)

        get_metadata_item_method = getattr(
            db, self.resource_metadata_get_item[self.resource_name])
        get_metadata_item_method(context, resource_id, key)

        delete_metadata_method = getattr(
            db, self.resource_metadata_delete[self.resource_name])
        delete_metadata_method(context, resource_id, key)
